use strict;
use Data::Dumper;
use Readonly;
use Test::More;
use DE_EPAGES::Core::API::Error qw ( ExistsError GetError );
use DE_EPAGES::Amazon::API::Cartridge;
use DE_EPAGES::Amazon::API::Object::PagerMWSProductItems;
use DE_EPAGES::Core::API::Script qw ( RunScript );
use DE_EPAGES::Database::API::Connection qw ( GetTestSite RunOnStore );
use DE_EPAGES::Object::API::XML::ObjectImportHandler qw ( ImportObjects );
use DE_EPAGES::Amazon::API::RemoteAdapterFactory;
use DE_EPAGES::Object::API::Factory qw (
    ExistsObjectByPath
    LoadObjectByPath
    LoadRootObject
);

Readonly my $ROOTPATH => '/Test';

sub TestPager {
    my $PortalConfig = LoadObjectByPath($ROOTPATH)->path('TestShopAmazon/PortalConfigs/AmazonTestConfig');
    my $RemoteAWS = DE_EPAGES::Amazon::API::RemoteAdapterFactory::NewRemoteMWSProduct( $PortalConfig );
    $RemoteAWS->setLogging(1);
    ok (my $Pager = testSearch( 'Deuter', $RemoteAWS ), 'search for "Deuter"');
    cmp_ok( scalar @{ $Pager->items() }, '>=', 1, 'search for "Deuter" returns at least one result' );
    ok(my $TestItem = $Pager->items()->[0], 'fetch first "Deuter"-Item');
    ok ($TestItem->get('ASIN'),'"Deuter"-Item contains a ASIN');
  SKIP: {
      my  @SkipVersions = qw(7.51.0 7.52.0);
      my $Version = DE_EPAGES::Amazon::API::Cartridge->new()->{'Version'};
      skip("testBrandOfFirstItems() in version $Version" ) if $Version ~~ @SkipVersions;
      ok(testBrandOfFirstItems($Pager,qr /Deuter/),'response of one of the first 3 items returns something with "Deuter"');
    }
    $Pager = testASINLookup( 'B001UF0QP8', $RemoteAWS );
    cmp_ok( scalar @{  $Pager->items() }, '>=', 1, 'search for Deuter backback ASIN returns at least one result' );
    like ($Pager->items()->[0]->get('Brand'),qr /Deuter/,'response returns something with "Deuter"');
    $Pager = testASINLookup( 'B0009VC8MI', $RemoteAWS );
    cmp_ok(  scalar @{ $Pager->items()->[0]->get('VariationsProducts')->{'DataSets'} }, '==', 0, 'search for Campingaz Kocher ASIN returns no variation product' );
    $Pager = testASINLookup( 'B00QFLCGYG', $RemoteAWS );
    cmp_ok( scalar @{ $Pager->items()->[0]->get('VariationsProducts')->{'DataSets'}->[0]->{'Data'} }, '>=', 2, 'search for Batman Arkham T-Shirt (variation product) returns at least 2 variations.' );
    $Pager = testASINLookup( 'B00XXXXXXX', $RemoteAWS );
    cmp_ok( scalar @{ $Pager->items() } , '==', 0, 'search for invalid ASIN returns 0 results.' );
}

sub testSearch {
    my ( $SearchString, $RemoteAWS ) = @_;

    my $Pager = DE_EPAGES::Amazon::API::Object::PagerMWSProductItems->new( {
        'ObjectID' => 1,
        'Remote'         => $RemoteAWS,
        'Parameters' => {
            'ASINs'          => undef,
            'UPCEAN'         => undef,
            'SearchString'   => $SearchString,
            'ProductName'    => $SearchString,
        }
    });

    $Pager->statistic;
    return $Pager;
}

sub testBrandOfFirstItems {
    my ( $Pager, $Brand) = @_;
    for(my $i=0; $i<3; $i++){
        my $Item = $Pager->items()->[$i];
        ## no critic qw( ProhibitPrivateSubroutineUse )
        return 1 if( (defined $Item) && $Item->_existsAttribute('Brand') && $Item->get('Brand') =~ $Brand );
    }
    return 0;
}

sub testASINLookup {
    my ( $ASIN, $RemoteAWS ) = @_;

    my $Pager = DE_EPAGES::Amazon::API::Object::PagerMWSProductItems->new( {
        'ObjectID' => 1,
        'Remote' => $RemoteAWS,
        'Parameters' => {
            'ASINs' => $ASIN,
            'UPCEAN' => undef,
        }
    });
    if ( defined( $Pager->{'XMLResponse'}->{'OperationRequest'}->{'Errors'}->{'Error'} )
        && $Pager->{'XMLResponse'}->{'OperationRequest'}->{'Errors'}->{'Error'}->[0]->{'Code'} eq 'AWS.TimestampExceedsTimeToLive' )
    {
        Test::More::diag('AWS.TimestampExceedsTimeToLive - wrong clock! ASIN search will fail.');
    }

    $Pager->page();
    return $Pager;
}

sub Main {
    RunOnStore(
        'Store' => GetTestSite(),
        'Sub'   => sub {

            # startUp
            ok( Init(), 'initialization' );

            # tests
            eval {
                TestPager();
                 };

            if ( ExistsError() ) {
                my $Error = GetError();
                fail( 'Error: ' . $Error->longMessage );
            }

            # shutDown
            ok( Cleanup(), 'cleanup' );
            done_testing();
        }
    );
}

sub Cleanup {
    return 1 if $ENV{'SKIP_CLEANUP'};
    LoadObjectByPath($ROOTPATH)->delete if ExistsObjectByPath($ROOTPATH);
    return 1;
}

sub Init {
    return 1 if ExistsObjectByPath($ROOTPATH) && $ENV{'SKIP_CLEANUP'};
    Cleanup();
    LoadRootObject()->insertChild( 'Object', { Alias => 'Test' } );
    ImportObjects( 'ImportForMWSProduct.xml', $ROOTPATH );
    return 1;
}

RunScript( Sub => \&Main );

1;
