#========================================================================================
# §package      DE_EPAGES::Core::API::Image
# §state        public
#----------------------------------------------------------------------------------------
# §description  This module provides image functionality.
#----------------------------------------------------------------------------------------
# §example      use DE_EPAGES::Core::API::Image qw( AddBorderToImage NewImage ...);
#========================================================================================
package DE_EPAGES::Core::API::Image;
use base Exporter;

use strict;
our @EXPORT_OK = qw(
  ScaleImage
  ResizeImage
  ResizeImageUsingImageRef
  NewImage
  Composite
  AddBorderToImage
  GetImageInfo
  TestImage
  TestImageFormat
  CreateIcon
  CorrectImage
  ConvertCmToPx
);

use Config;
use DateTime;
use Image::ExifTool qw(ImageInfo);
use Image::Magick;
use Image::Size qw(imgsize);
use List::Util qw( max );
use DE_EPAGES::Core::API::Constants;
use DE_EPAGES::Core::API::Error qw( Error GetError ExistsError );
use DE_EPAGES::Core::API::File
  qw( SetAccessRights GetAccessRights fsplit ConvertPathChars ExistsFile DeleteFile MoveFile GetFileInfo );
use DE_EPAGES::Core::API::TestParameter qw( TestString TestInteger TestObject );
use DE_EPAGES::Core::API::Warning qw( Warning );
use POSIX;

# important to switch off the cache!
$Image::Size::NO_CACHE = 1;

use constant {
    MAX_IMAGE_QUALITY => 100,
    ONE_MEGA_PIXEL => 1048576
};

#ERRORDEFINITION
sub ErrorDefinition () {
    return {
        'INVALID_IMAGE_FILE' =>
          '#File doesn\'t seem to be a valid image file: #ImageMagickError',
        'IMAGE_EXCEEDS_MEGAPIXELS' => 'The image dimensions of #File exceed '
          . MAX_IMAGE_MEGAPIXELS
          . ' megapixels',
        'IMAGE_MAGICK' => 'ImageMagick#Function (#File): #Message',
        'UNSUPPORTED_IMAGE_FORMAT' =>
          '#Parameter is not a supported image file (#File)',
    };
}

#ERRORDEFINITION END

my %IMAGE_FORMATS = map { '.' . $_ => 1 } qw(gif jpg jpeg png ico);



#========================================================================================
# §function     TestImageFormat
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       TestImageFormat($Parameter, $FilePath);
# §example      TestImageFormat('SourceFile', $SourceFile);
#----------------------------------------------------------------------------------------
# §description  Throws the error UNSUPPORTED_IMAGE_FORMAT if the file name has an
#               extension that is not supported.
#               Supported extensions are: bmp gif jpg jpeg png.
#----------------------------------------------------------------------------------------
# §input        $Parameter | parameter name | string
# §input        $FilePath | fully qualified file name | string
#========================================================================================
sub TestImageFormat {
    my ( $Parameter, $FilePath ) = @_;

    TestString( 'Parameter', $Parameter );
    TestString( 'FilePath',  $FilePath );

    my ( undef, undef, $Extension ) = fsplit($FilePath);
    unless ( exists $IMAGE_FORMATS{ lc($Extension) } ) {
        Error(
            'UNSUPPORTED_IMAGE_FORMAT',
            {
                Parameter => $Parameter,
                Extension => $Extension,
                File      => $FilePath
            }
        );
    }

    return;
}

#========================================================================================
# §function     TestImage
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       TestImage($FilePath);
# §example      TestImage('SourceFile.jpg');
#----------------------------------------------------------------------------------------
# §description  Throws the error INVALID_IMAGE_FILE if the file is not an image, that is
#               understood by Image::Magick
#----------------------------------------------------------------------------------------
# §input        $FilePath | fully qualified file name | string
#========================================================================================
sub TestImage {
    my ($FilePath) = @_;
    TestString( 'FilePath', $FilePath );

    Error( 'FILE_NOT_FOUND', { 'FileName' => $FilePath } ) unless -f $FilePath;

    my $ImageInfo = ImageInfo($FilePath);
    if ( $ImageInfo->{'ImageHeight'} && $ImageInfo->{'ImageWidth'} ) {
        if (
            (
                $ImageInfo->{'ImageHeight'} *
                $ImageInfo->{'ImageWidth'} / ONE_MEGA_PIXEL
            ) > MAX_IMAGE_MEGAPIXELS
          )
        {    ## no critic qw(ProhibitMagicNumbers)
            Error(
                'IMAGE_EXCEEDS_MEGAPIXELS',
                {
                    'File'             => $FilePath,
                    'ImageMagickError' => 'The image dimensions exceed '
                      . MAX_IMAGE_MEGAPIXELS
                      . ' megapixels.',
                    'MegaPixelLimit' => MAX_IMAGE_MEGAPIXELS,
                }
            );
            return;
        }
    }

    my $Image  = Image::Magick->new;
    my $Result = $Image->Read($FilePath);
    if ($Result) {
        Error(
            'INVALID_IMAGE_FILE',
            {
                'File'             => $FilePath,
                'ImageMagickError' => $Result,
            }
        );
    }

    return;
}

#========================================================================================
# §function     ScaleImage
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       ScaleImage( $Source, $Target, $ScaleSize, $ScaleHeight, $Quality );
# §example      ScaleImage( "/tmp/image.gif", "/tmp/image2.gif", 80 );
#----------------------------------------------------------------------------------------
# §description  Creates a new image where the maximal width or height is $ScaleSize.
#----------------------------------------------------------------------------------------
# §input        $Source | source file | string
# §input        $Target | target file | string
# §input        $ScaleSize | maximal size (width or height) of new image | int
# §input        $ScaleHeight | if ScaleHeight == 1 then scale height only else scale
#               width and height (optional)| bit
# §input        $Quality | JPEG quality 0 - 100, DEFAULT_JPEG_QUALITY if undefined, optional | integer
#========================================================================================
sub ScaleImage {
    my ( $Source, $Target, $ScaleSize, $ScaleHeight, $Quality ) = @_;
    TestImageFormat( 'Source', $Source );
    TestString( 'Target', $Target );
    TestInteger( 'ScaleSize', $ScaleSize, 1 );
    if ( defined $Quality ) {
        TestInteger( 'Quality', $Quality, 0, MAX_IMAGE_QUALITY )
          ;    ## no critic qw(ProhibitMagicNumbers)
    }
    else {
        $Quality = DEFAULT_JPEG_QUALITY;
    }
    Error( 'FILE_NOT_FOUND', { 'FileName' => $Source } ) unless -f $Source;

    my $Image      = Image::Magick->new;
    my $ImageError = $Image->Read($Source);
    Error( 'IMAGE_MAGICK',
        { 'Function' => 'Read', 'File' => $Source, 'Message' => $ImageError } )
      if $ImageError;

    my $ImageHeight = $Image->Get('height');
    my $ImageWidth  = $Image->Get('width');

    my $Scale =
      ( ( $ImageHeight > $ImageWidth or $ScaleHeight )
        ? $ImageHeight
        : $ImageWidth ) / $ScaleSize;

    my $NewWidth  = sprintf( '%.0f', $ImageWidth / $Scale );
    my $NewHeight = sprintf( '%.0f', $ImageHeight / $Scale );

    ResizeImage( $Source, $Target, $NewWidth, $NewHeight, $Quality );
}

#========================================================================================
# §function     ResizeImage
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       ResizeImage( $Source, $Target, $ScaleWidth, $ScaleHeight, $Quality );
# §example      ResizeImage( "/tmp/image.gif", "/tmp/image2.gif", 320, 240, 80 );
#----------------------------------------------------------------------------------------
# §description  Creates a new image where the maximal width is $ScaleWidth and maximal
#               height is $ScaleHeight.
#----------------------------------------------------------------------------------------
# §input        $Source | source file | string
# §input        $Target | target file | string
# §input        $ScaleWidth | maximal width of new image (optional) | int
# §input        $ScaleHeight | maximal height of new image (optional) | int
# §input        $Quality | JPEG quality 0 - , (default: see DEFAULT_JPEG_QUALITY) | integer
#========================================================================================
sub ResizeImage {
    my ( $Source, $Target, $ScaleWidth, $ScaleHeight, $Quality ) = @_;

    # check if supported image format [EPG-10204]
    TestImageFormat( 'Source', $Source );

    if ( defined $Quality ) {
        TestInteger( 'Quality', $Quality, 0, MAX_IMAGE_QUALITY )
          ;    ## no critic qw(ProhibitMagicNumbers)
    }
    else {
        $Quality = DEFAULT_JPEG_QUALITY;
    }
    Error( 'FILE_NOT_FOUND', { 'FileName' => $Source } ) unless -f $Source;

    my $Image = Image::Magick->new;

    my $ImageError = $Image->Read($Source);
    Error( 'IMAGE_MAGICK',
        { 'Function' => 'Read', 'File' => $Source, 'Message' => $ImageError } )
      if $ImageError;

    # get image width and height
    my $ImageHeight = $Image->Get('height');
    my $ImageWidth  = $Image->Get('width');

    # possible image magick cant read width/height
    if ( not defined $ImageWidth or not defined $ImageHeight ) {
        my $ImageInfo = ImageInfo($Source);
        if (   defined( $ImageInfo->{'ImageHeight'} )
            && defined( $ImageInfo->{'ImageWidth'} ) )
        {

            # try to determine image dimensions via exif
            $ImageHeight = $ImageInfo->{'ImageHeight'};
            $ImageWidth  = $ImageInfo->{'ImageWidth'};
        }
        else {
            Error(
                'UNKNOWN_IMAGE_SIZE',
                {
                    'File'   => $Source,
                    'Width'  => $ImageWidth,
                    'Height' => $ImageHeight
                }
            );
        }
    }

    my $Error =
      _setResolution( $Image, $Source, $Target, $ImageWidth, $ImageHeight,
        $ScaleWidth, $ScaleHeight, $Quality );

    undef $Image;
    _handleImageError($Error, $Target);

    SetAccessRights( $Target, GetAccessRights($Source) );
    return;
}

#========================================================================================
# §function     _setResolution
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       _setResolution( $SourceImage, $Source, $Target, $ImageWidth, $ImageHeight, $ScaleWidth, $ScaleHeight, $Quality );
# §example      _setResolution(
#                   $WorkImage,
#                   $Source,
#                   $Path.'/'.$Target,
#                   $ImageWidth,
#                   $ImageHeight,
#                   $hParams->{'SCALE_WIDTH_'.$Size},
#                   $hParams->{'SCALE_HEIGHT_'. $Size},
#                   $Quality
#               );
#----------------------------------------------------------------------------------------
# §description  Scales the image.
#----------------------------------------------------------------------------------------
# §input        $SourceImage | image magick object | object
# §input        $Source | source file | string
# §input        $Target | target file | string
# §input        $ImageWidth | width of source image | int
# §input        $ImageHeight | height of new source image | int
# §input        $ScaleWidth | maximal width of new image (optional) | int
# §input        $ScaleHeight | maximal height of new image (optional) | int
# §input        $Quality | JPEG quality 0 - 100, (default DEFAULT_JPEG_QUALITY) | integer
#----------------------------------------------------------------------------------------
# §return       $Error | hash with errors and function | ref.hash
#========================================================================================
sub _setResolution {

    my (
        $SourceImage, $Source,     $Target,      $ImageWidth,
        $ImageHeight, $ScaleWidth, $ScaleHeight, $Quality
    ) = @_;

    # define scale
    my $Scale = 0;
    if ( $ScaleWidth && $ScaleHeight ) {
        $Scale = max( $ImageWidth / $ScaleWidth, $ImageHeight / $ScaleHeight );
    }
    elsif ($ScaleHeight) {
        $Scale = $ImageHeight / $ScaleHeight;
    }
    elsif ($ScaleWidth) {
        $Scale = $ImageWidth / $ScaleWidth;
    }    # else Scale = 0 copy

    my $ImageError;

    my $Function = 'Write';
    # only scale if orig. is larger than $ScaleSize
    if ( $Scale < 1 ) {
        $ImageError = $SourceImage->Write($Target) unless $Source eq $Target;
    } else {
        $Function = 'Scale';

        # Fix issues with animated gifs that use sub-layering with
        # background layers and transparency.
        if ( defined $SourceImage->[1] ) {    # anigif = multi-images
            $SourceImage = $SourceImage->Coalesce();
        }

        # For each subimage (think of animated gifs)
        for ( my $x = 0 ; $SourceImage->[$x] && !$ImageError ; $x++ ) {
            $ImageError =
              $SourceImage->[$x]->Thumbnail(
                'geometry' => sprintf( '%.0f', $ImageWidth / $Scale ) . 'x'
                  . sprintf( '%.0f', $ImageHeight / $Scale ), );
        }

   
        # To fix UNITY-6490 we commented this workaround out. It should be reimplemeted if the bug in imagemagick still exists
        # workaround due to imagemagick bug
        # my $Type = $SourceImage->Get('type');
        # if ( $Type ne 'PaletteMatte' ) {
        #     # prevent change of 8bit pngs to 24bit pngs
        #     $ImageError = $SourceImage->Set( 'type' => $Type )
        #         unless ($ImageError && $ImageError !~ IMAGE_MAGICK_WARNING);
        # }



        # do not change the compression level
        $ImageError = $SourceImage->Set( 'quality' => $Quality )
            unless ($ImageError && $ImageError !~ IMAGE_MAGICK_WARNING);
        $ImageError = $SourceImage->Write($Target)
            unless ($ImageError && $ImageError !~ IMAGE_MAGICK_WARNING);
    }

    # write EXIF information to resized image
    my $ExifTool = GetImportantMetaData($Source, $SourceImage);
    $ExifTool->WriteInfo($Target);

    return { 'Error' => $ImageError, 'Function' => $Function };
}


#========================================================================================
# §function     _handleImageError
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       _handleImageError( $Error, $Target );
#----------------------------------------------------------------------------------------
# §description  Checks, if there is an error or an warning and throws the related epages
#               exception.
#----------------------------------------------------------------------------------------
# §input        $Error | hash with errors and function (see _setResolution) | ref.hash
# §input        $Target | target file | string
#========================================================================================
sub _handleImageError {
    my ($Error, $Target) = @_;

    my $ImageError = $Error->{'Error'};

    return unless $ImageError;

    my $Function   = $Error->{'Function'};
    my $hVars = {
        'Function' => $Function,
        'File'     => $Target,
        'Message'  => $ImageError
    };

    # only throw errors that are not just warnings (UNITY-4559)
    if ($ImageError =~ IMAGE_MAGICK_WARNING) {
        Warning($ImageError, $hVars);
    } else {
        unlink $Target;
        Error('IMAGE_MAGICK', $hVars);
    }

    return;
}

#========================================================================================
# §function     ResizeImageUsingImageRef
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       ResizeImageUsingImageRef( $Image, $Source, $Target, $ImageWidth, $ImageHeight,
#                                         $ScaleWidth, $ScaleHeight, $Quality );
# §example      ResizeImageUsingImageRef(
#                   $WorkImage,
#                   $Source,
#                   $Path.'/'.$Target,
#                   $ImageWidth,
#                   $ImageHeight,
#                   $hParams->{'SCALE_WIDTH_'.$Size},
#                   $hParams->{'SCALE_HEIGHT_'. $Size},
#                   $Quality
#               );
#----------------------------------------------------------------------------------------
# §description  Like the ResizeImage method, but uses an already open Image::Magick
#               object reference instead of opening the source file. This can be useful
#               when applying more than operation to the image, multiple resize operations
#               for example.
#----------------------------------------------------------------------------------------
# §input        $SourceImage | image magick object | object
# §input        $Source | source file | string
# §input        $Target | target file | string
# §input        $ImageWidth | width of source image | int
# §input        $ImageHeight | height of new source image | int
# §input        $ScaleWidth | maximal width of new image (optional) | int
# §input        $ScaleHeight | maximal height of new image (optional) | int
# §input        $Quality | JPEG quality 0 - 100, (default DEFAULT_JPEG_QUALITY) | integer
#========================================================================================
sub ResizeImageUsingImageRef {
    my (
        $SourceImage, $Source,     $Target,      $ImageWidth,
        $ImageHeight, $ScaleWidth, $ScaleHeight, $Quality
    ) = @_;

    if ( defined $Quality ) {
        TestInteger( 'Quality', $Quality, 0, MAX_IMAGE_QUALITY )
          ;    ## no critic qw(ProhibitMagicNumbers)
    }
    else {
        $Quality = DEFAULT_JPEG_QUALITY;
    }

    my $Error = _setResolution(
        $SourceImage, $Source,     $Target,      $ImageWidth,
        $ImageHeight, $ScaleWidth, $ScaleHeight, $Quality
    );

    _handleImageError($Error, $Target);

    SetAccessRights( $Target, GetAccessRights($Source) );
    return;
}

#========================================================================================
# §function     NewImage
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       NewImage( $Target, $hParams);
# §example      NewImage( "/tmp/image.gif", {'Height' => 150,
#                                            'Width'  => 150,
#                                            'BGColor'=>'#FFFFFF'});
#----------------------------------------------------------------------------------------
# §description  Create a new image with the given params.
#----------------------------------------------------------------------------------------
# §input        $Target | source file | string
# §input        $hParams | possible params are:
#                         <ul>
#                           <li>Height - target image height</li>
#                           <li>Width  - target image width</li>
#                           <li>BGColor (optional) - target image background color</li>
#                         </ul> | ref.hash.string
#========================================================================================
sub NewImage {
    my ( $Target, $hParams ) = @_;

    my $ImageHeight = $hParams->{'Height'};
    my $ImageWidth  = $hParams->{'Width'};

    my $ImageBGColor =
      $hParams->{'BGColor'} ? $hParams->{'BGColor'} : '#ffffff';

    TestInteger( 'ImageHeight', $ImageHeight, 1 );
    TestInteger( 'ImageWidth',  $ImageWidth,  1 );

    my $Image = Image::Magick->new();
    my $IErr = $Image->Set( size => "$ImageWidth" . 'x' . "$ImageHeight" );
    Error( 'IMAGE_MAGICK', { 'Function' => 'Set', 'Message' => $IErr } )
      if $IErr;

    $IErr = $Image->ReadImage();
    $IErr = $Image->Colorize( 'fill' => $ImageBGColor );
    Error( 'IMAGE_MAGICK', { 'Function' => 'Colorize', 'Message' => $IErr } )
      if $IErr;

    $IErr = $Image->Write($Target);
    Error( 'IMAGE_MAGICK',
        { 'Function' => 'Write', 'File' => $Target, 'Message' => $IErr } )
      if $IErr;

    undef $Image;
    return;
}

#========================================================================================
# §function     Composite
# §state        public (obsolete! use AddBorder() )
#----------------------------------------------------------------------------------------
# §syntax       Composite( $Source,$Target,$hParams);
# §example      Composite( "/tmp/image.gif","/tmp/imagenew.gif", {'Gravity' => 'Center',});
#----------------------------------------------------------------------------------------
# §description  Overlay the target image with the source image
#               - the gravity describes the alignment of the overlayed image
#
#               Obsolete function - use AddBorder()
#----------------------------------------------------------------------------------------
# §input        $Source | source file | string
# §input        $Target | target file | string
# §input        $hParams | possible params are:
#                         <ul>
#                           <li>Gravity (optional default: Center) - gravity</li>
#                         </ul> | ref.hash.string
#========================================================================================
sub Composite {
    my ( $Source, $Target, $hParams ) = @_;
    TestString( 'Source', $Source );
    TestImageFormat( 'Source', $Source );
    TestString( 'Target', $Target );

    Error( 'FILE_NOT_FOUND', { 'FileName' => $Source } ) unless -f $Source;
    Error( 'FILE_NOT_FOUND', { 'FileName' => $Target } ) unless -f $Target;

    my $Gravity = $hParams->{'Gravity'} ? $hParams->{'Gravity'} : 'Center';

    my $SImage = Image::Magick->new();
    my $TImage = Image::Magick->new();

    my $IErr = $SImage->Read($Source);
    Error( 'IMAGE_MAGICK',
        { 'Function' => 'Read', 'File' => $Source, 'Message' => $IErr } )
      if $IErr;

    $IErr = $TImage->Read($Target);
    Error( 'IMAGE_MAGICK',
        { 'Function' => 'Read', 'File' => $Target, 'Message' => $IErr } )
      if $IErr;

    my $SClone = $TImage->Clone();
    undef $TImage;

    $SClone->Composite(
        image   => $SImage,
        compose => 'over',
        gravity => $Gravity
    );

    Error( 'IMAGE_MAGICK', { 'Function' => 'Composite', 'Message' => $IErr } )
      if $IErr;
    undef $SImage;
    $IErr = $SClone->Write($Target);
    Error( 'IMAGE_MAGICK',
        { 'Function' => 'Write', 'File' => $Target, 'Message' => $IErr } )
      if $IErr;
    undef $SClone;
    SetAccessRights( $Target, GetAccessRights($Source) );
    return;
}

#========================================================================================
# §function     _AddBorder
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       _AddBorder( $Source,$Target,$hParams);
# §example      _AddBorder( "/tmp/image.gif", {'Height' => 150,
#                                             'Width'  => 150,
#                                             'BGColor'=>'#FFFFFF',
#                                             'CalculateDiff'=>1,
#----------------------------------------------------------------------------------------
# §description  Add a border to the source image and save this image as target
#               - the CalculateDiff flag is used to calculate the border for the image
#                 to get the target image size of Height and Width
#
#               - Attention: Height value is added on both sides
#               - Attention: Width value is added on both sides
#----------------------------------------------------------------------------------------
# §input        $Source | source file | string
# §input        $Target | target file | string
# §input        $hParams | possible params are:
#                         <ul>
#                           <li>CalculateDiff (optional default: 0) - CalculateDiff</li>
#                         </ul> | ref.hash.string
#========================================================================================
sub _AddBorder {
    my ( $Source, $Target, $hParams ) = @_;
    TestString( 'Source', $Source );
    TestImageFormat( 'Source', $Source );
    TestString( 'Target', $Target );

    my $ImageHeight = $hParams->{'Height'};
    my $ImageWidth  = $hParams->{'Width'};
    my $ImageBGColor =
      $hParams->{'BGColor'} ? $hParams->{'BGColor'} : '#ffffff';

    TestInteger( 'ImageHeight', $ImageHeight, 1 );
    TestInteger( 'ImageWidth',  $ImageWidth,  1 );
    Error( 'FILE_NOT_FOUND', { 'FileName' => $Source } ) unless -f $Source;

    my $SImage = Image::Magick->new();
    my $IErr   = $SImage->Read($Source);

    Error( 'IMAGE_MAGICK',
        { 'Function' => 'Read', 'File' => $Source, 'Message' => $IErr } )
      if $IErr;

    my ( $OriginalWidth, $OriginalHeight ) = ( $ImageWidth, $ImageHeight );

    # get the size of the original and calculate the Border dimensions...
    if ( $hParams->{'CalculateDiff'} ) {
        ( $OriginalHeight, $OriginalWidth ) = $SImage->Get( 'height', 'width' );
        $ImageHeight = ( $ImageHeight - $OriginalHeight ) / 2;
        $ImageWidth  = ( $ImageWidth - $OriginalWidth ) / 2;
    }

    # do not allow negative sizes
    $ImageWidth  = 0 if $ImageWidth < 0;
    $ImageHeight = 0 if $ImageHeight < 0;

    my $Dispose = $SImage->Get('dispose');
    if ( $Dispose eq 'None' ) {
        $IErr = $SImage->[0]->Border(
            width       => $ImageWidth,
            height      => $ImageHeight,
            bordercolor => $ImageBGColor,
        );
        Error( 'IMAGE_MAGICK',
            { 'Function' => 'Border', 'File' => $Target, 'Message' => $IErr } )
          if $IErr;
        for ( my $i = 1 ; $i < scalar @$SImage ; $i++ ) {
            $IErr = $SImage->[$i]->Border(
                width       => $ImageWidth,
                height      => $ImageHeight,
                bordercolor => 'transparent',
            );
            Error(
                'IMAGE_MAGICK',
                {
                    'Function' => 'Border',
                    'File'     => $Target,
                    'Message'  => $IErr
                }
            ) if $IErr;
        }
    }
    else {
        for ( my $i = 0 ; $i < scalar @$SImage ; $i++ ) {
            $IErr = $SImage->[$i]->Extent(
                width      => $OriginalWidth,
                height     => $OriginalHeight,
                background => $ImageBGColor,
            );
            Error(
                'IMAGE_MAGICK',
                {
                    'Function' => 'Extent',
                    'File'     => $Target,
                    'Message'  => $IErr
                }
            ) if $IErr;
            $IErr = $SImage->[$i]->Border(
                width       => $ImageWidth,
                height      => $ImageHeight,
                bordercolor => $ImageBGColor,
            );
            Error(
                'IMAGE_MAGICK',
                {
                    'Function' => 'Border',
                    'File'     => $Target,
                    'Message'  => $IErr
                }
            ) if $IErr;
        }
    }
    $IErr = $SImage->Write($Target);
    Error( 'IMAGE_MAGICK',
        { 'Function' => 'Write', 'File' => $Target, 'Message' => $IErr } )
      if $IErr;

    undef $SImage;
    SetAccessRights( $Target, GetAccessRights($Source) );
    return;
}

#========================================================================================
# §function     AddBorderToImage
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       AddBorderToImage( $Target, $hParams);
# §example      AddBorderToImage( "/tmp/image.gif", {'Height' => 150,
#                                                    'Width'  => 150,
#                                                    'BGColor'=>'#FFFFFF'});
#----------------------------------------------------------------------------------------
# §description  Add a border with the given params to the image.
#               - Add a border to the image. Then the new was renamed to the target image.
#----------------------------------------------------------------------------------------
# §input        $Target | source file | string
# §input        $hParams | possible params are:
#                         <ul>
#                           <li>Height - target image height</li>
#                           <li>Width - target image width</li>
#                           <li>BGColor (optional) - target image background color</li>
#                         </ul> | ref.hash.string
#========================================================================================
sub AddBorderToImage {

    my ( $Target, $hParams ) = @_;
    TestString( 'Target', $Target );
    TestImageFormat( 'Target', $Target );

    Error( 'FILE_NOT_FOUND', { 'FileName' => $Target } ) unless -f $Target;
    my ( $Path, $FileName, $Extension ) = fsplit($Target);
    $FileName = ConvertPathChars($FileName);
    my $TMPTarget = $Path . '/' . $FileName . '_TMP' . $Extension;
    $hParams->{'CalculateDiff'} = 1;
    _AddBorder( $Target, $TMPTarget, $hParams );

    # test if the target file really exists... (avoid nfs problems)
    if ( not ExistsFile($TMPTarget) ) {
        sleep(1);
        if ( not ExistsFile($TMPTarget) ) {
            sleep(1);
        }
    }
    DeleteFile($Target);
    MoveFile( $TMPTarget, $Target );
    return;
}

#========================================================================================
# §function     GetImageInfo
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $hImageInfo = GetImageInfo($FileName, $WithExif);
#----------------------------------------------------------------------------------------
# §description  Returns ref.hash with mode, uid and gid for FileName (only unix).
#----------------------------------------------------------------------------------------
# §input        $FileName | file name to get informations from | string
# §input        $WithExif | as exif data getting is very expensive, you can use this
#               parameter to decide, if this function should return the exif information
#               (optional) | boolean
# §return       $hImageInfo | flags
#                   <ul>
#                       <li>Name            - filename with Path
#                       <li>FileName        - filename without Path
#                       <li>Extension       - extension of file
#                       <li>Path            - path of file
#                       <li>Size            - file size in byte
#                       <li>LastAccess      - DateTime Object
#                       <li>LastChange      - DateTime Object
#                       <li>Height          - Height of image
#                       <li>Width           - Width of image
#                       <li>Mimetype        - Mimetype "image/extension"
#                       <li>Exif            - hash with Exif tags (see $WithExif)
#                   </ul>
#               | ref.hash
#========================================================================================
sub GetImageInfo($;$) {
    my ( $FileName, $WithExif ) = @_;
    $WithExif = 0 unless defined $WithExif;
    my $hInfo = GetFileInfo($FileName);
    if ( exists $IMAGE_FORMATS{ lc $hInfo->{'Extension'} } ) {
        TestImageFormat( 'FileName', $FileName );
        ( $hInfo->{'Width'}, $hInfo->{'Height'} ) = imgsize($FileName);
        if ($WithExif) {
            $hInfo->{'Exif'}     = ImageInfo($FileName);

            $hInfo->{'Mimetype'} = $hInfo->{'Exif'}->{'MIMEType'}
              if ( exists $hInfo->{'Exif'}->{'MIMEType'} );

            # Exif datetime fields:
            # DateTimeOriginal  # currently not needed
            # ModifyDate        # currently not needed
            # MetadataDate      # currently not needed
            # ProfileDateTime   # currently not needed
            # FileModifyDate    # currently not needed
            # CreateDate        # used by imagegallery
            my @AllowedExifDateTimeFields = qw (
                CreateDate
            );

            # convert all exif datetime values to a DateTime object
            foreach my $Key (
                grep {
                    defined $hInfo->{'Exif'}->{$_}
                      and ref $hInfo->{'Exif'}->{$_} eq ''
                } %{ $hInfo->{'Exif'} }
              )
            {
                next unless $Key ~~ @AllowedExifDateTimeFields;
                my $Value = $hInfo->{'Exif'}->{$Key};

                # replace time string by datetime objects
                if ( $Value =~ /^(\d{4})      # Year with 4 digits e.g. 1945
                                \D+(\d{1,2})  # a separator and the month with 2 digits e.g .12
                                \D+(\d{1,2})  # a separator and the day with 2 digits e.g .24
                                \D+(\d{1,2})  # a separator and the hours with 2 digits e.g .13
                                \D+(\d{1,2})  # a seperator and the minutes with 2 digits e.g. .15
                                \D+(\d{1,2})$ # a seperator and the secondes with 2 digits e.g .00
                               /x )
                {
                    # assign new datetime object only if valid value can be found
                    $hInfo->{'Exif'}->{$Key} = eval { ## no critic qw(RequireExistsError) # ignore invalid DateTime values
                        DateTime->new(
                            'year'   => $1,
                            'month'  => $2,
                            'day'    => $3,
                            'hour'   => $4,
                            'minute' => $5,
                            'second' => $6
                        );
                    };
                }
            }
        }
    }
    return $hInfo;
}

#========================================================================================
# §function     CreateIcon
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       CreateIcon($Image, $Icon);
# §example      CreateIcon($Shop->get('PublicPath').'/'.$Shop->get('Logo'), $Shop->get('PublicPath').'/'.'favicon.ico');
#----------------------------------------------------------------------------------------
# §description  Creates icon from image, the icon includes 3 (16x16,32x32,64x64) sizes.
#----------------------------------------------------------------------------------------
# §input        $Image | source image file name | string
# §input        $Icon | target icon file name | string
#========================================================================================
sub CreateIcon {
    my ( $Source, $Target ) = @_;

    my $Image = Image::Magick->new;
    $Image->Read( $Source, $Source, $Source );
    return unless defined $Image->[0];

    $Image->[0]->Resize( width => 16, height => 16, blur => 0.8 );
    $Image->[1]->Resize( width => 32, height => 32, blur => 0.8 );
    $Image->[2]->Resize( width => 64, height => 64, blur => 0.8 );
    $Image->Write( 'filename' => $Target, 'compression' => 'None' );
    SetAccessRights( $Target, GetAccessRights($Source) );

    return;
}

#========================================================================================
# §function    CorrectImage
# §state       public
#----------------------------------------------------------------------------------------
# §syntax      CorrectImage( $File );
# §syntax      CorrectImage( undef, $Image );
# §example     CorrectImage( "/tmp/image.gif" );
#----------------------------------------------------------------------------------------
# §description Converts the image color space to RGB if it is not yet an RGB image.
#               Implemented for EPG-11560 to fix CMYK images. To save an additional
#               $Image->Read($File) you can also supply a valid image magick object.
#               Also strips color profiles and comments from the image.
#----------------------------------------------------------------------------------------
# §input       $File | full file path and name for a image file to be fixed | string
# §input       $Image | (optional) image magick object.
#                   If this parameter is defined, then $File is ignored. | object
#========================================================================================
sub CorrectImage {
    my ( $File, $Image ) = @_;

    my $UseRef = $Image ? 1 : 0;

    if( $UseRef ) {
        TestObject( 'Image', $Image, 'Image::Magick' );
    }
    else {
        TestString( 'File', $File );

        # check if supported image format
        TestImageFormat( 'File', $File );
        if ( ExistsError() ) {
            my $Error = GetError();
            return if $Error->code eq 'UNSUPPORTED_IMAGE_FORMAT';
            Error( $Error ); # report unexpected errors
        }

        # open image file
        $Image = Image::Magick->new;
        $Image->Read($File);
    }

    # Remove meta data and save the data to keep
    my $ExifTool = GetImportantMetaData($File, $Image);

    # check if current color space is unequal to RGB
    my $CurrentColorSpace = $Image->Get('colorspace');
    if ( $CurrentColorSpace !~ /RGB/g ) {
        # change color space to rgb
        $Image->Set( colorspace => 'RGB' );
        $Image->Quantize( colorspace => 'RGB' );
    }

    # Coalesce, if animated image.
    if(defined $Image->[1]) {
        $Image = $Image->Coalesce();
    }

    $Image->Write($File) if !$UseRef;
    $ExifTool->WriteInfo($File);
}

#========================================================================================
# §function     GetImportantMetaData
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       GetImportantMetaData( $File, $Image );
#----------------------------------------------------------------------------------------
# §description  Removes most of the Metadata from an image and returns
#                the filtered data.
#----------------------------------------------------------------------------------------
# §input        $File      | full file path and name for a image file to be fixed | string
# §input        $Image     | The image of the file | image magick object.
# §return       $ExifTool  | The filtered meta data | ExifTool object
#========================================================================================
sub GetImportantMetaData {
    my ( $File, $Image ) = @_;

    my $MetaData = ImageInfo($File);
    my $ExifTool = Image::ExifTool->new;

    # We have to remember the original meta data from the image
    my $Artist          = $MetaData->{'Artist'};
    my $Rights          = $MetaData->{'Rights'};
    my $Copyright       = $MetaData->{'Copyright'};
    my $ModifyDate      = $MetaData->{'ModifyDate'};
    my $Software        = $MetaData->{'Software'};
    my $Description     = $MetaData->{'Description'};
    my $CreatorTool     = $MetaData->{'CreatorTool'};
    my $Orientation     = $MetaData->{'Orientation'};

    # Now remove all meta data from the given image, if the image is NO *.png
    my ( undef, undef, $Extension ) = fsplit($File);
    if(!(lc($Extension) eq ".png")) {
        $Image->Strip();
    }

    # Readd the saved meta data
    $ExifTool->SetNewValue(Artist => $Artist) if defined $Artist;
    $ExifTool->SetNewValue(Rights => $Rights) if defined $Rights;
    $ExifTool->SetNewValue(Copyright => $Copyright) if defined $Copyright;
    $ExifTool->SetNewValue(ModifyDate => $ModifyDate) if defined $ModifyDate;
    $ExifTool->SetNewValue(Software => $Software) if defined $Software;
    $ExifTool->SetNewValue(Description => $Description) if defined $Description;
    $ExifTool->SetNewValue(CreatorTool => $CreatorTool) if defined $CreatorTool;
    $ExifTool->SetNewValue(Orientation => $Orientation) if defined $Orientation;

    return $ExifTool;
}

#========================================================================================
# §function     ConvertCmToPx
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->ConvertCmToPx($DimensionCm)
#----------------------------------------------------------------------------------------
# §description  Converts a dimension from cm to px and rounds down.
#----------------------------------------------------------------------------------------
# §input        $DimensionCm | dimension in cm | float
# §input        $Density | image density (optional) | integer
# §return       $DimensionPixels | dimension in px | integer
#========================================================================================
sub ConvertCmToPx {
    my ($DimensionCm, $Density) = @_;
    my $ImageDensity = $Density // DEFAULT_DENSITY;

    # convert to px
    my $DimensionPixels = ($DimensionCm / INCH_IN_CM) * $ImageDensity;
    # round down
    return floor($DimensionPixels);
}

1;
