#========================================================================================
# package      DE_EPAGES::Coupon::API::Attributes::LineItemCoupon_Ext
# state        private
# base         DE_EPAGES::Object::API::BaseAttribute
# description  Attribute interface for LineItemCoupon.
#========================================================================================
package DE_EPAGES::Coupon::API::Attributes::LineItemCoupon_Ext;
use base qw( DE_EPAGES::Object::API::BaseAttribute );

use strict;
use DE_EPAGES::Core::API::PerlTools qw ( ElementOf);

#========================================================================================
# function     getAttributes
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hValues = $Package->getAttributes( $Object, $aNames );
# example      # never call getAttributes directly!
#               $Value = $Object->get( $AttributeName );
#----------------------------------------------------------------------------------------
# description  Returns attribute values of an object.
#               Used by <function DE_EPAGES::Object::Object::Object::get>
#----------------------------------------------------------------------------------------
# input        $Object | object | object
# input        $aNames | list of attribute names | ref.array
# return       $hValues | attribute name => value hash | ref.hash
#========================================================================================
sub getAttributes {
    my $self = shift;
    my ($Object, $aNames) = @_;
    my $hInfo = {};
    foreach my $Name (@$aNames) {
        if( $Name eq 'ValidationErrors' ) {
            $hInfo->{$Name} = $self->getValidationErrors($Object);
        }
    }
    return $hInfo;
}

sub getValidationErrors {
    my $self = shift;
    my ($LineItemCoupon) = @_;
    my @ErrorMessages;

    my $CouponCode = $LineItemCoupon->get('CouponCode');
    return [] unless defined $CouponCode;

    my $CouponCampaign = $CouponCode->get('CouponCampaign');
    my $LineItemsSubTotal = $LineItemCoupon->container->get('LineItemsSubTotalDiscountable');
    my $CouponDiscountSum = 0;
    $CouponDiscountSum += $_ foreach
        grep { defined $_ }
        map { $_->get('CouponDiscount') }
        grep { $_->instanceOf('LineItemProduct') }
        @{$LineItemCoupon->container->get('LineItems')};
    $LineItemsSubTotal += $CouponDiscountSum;
    if ($LineItemsSubTotal < $CouponCampaign->get('MinLineItemsSubTotal') ) {
        push (@ErrorMessages , { 'Message' => 'LineItemsSubTotalTooSmall'});
    }

    if (not $LineItemCoupon->get('LineItemContainer')->parent->instanceOf('CustomerOrder')){
        push @ErrorMessages, { 'Message' => 'CouponIsInvalidated'} if $CouponCode->get('IsInvalidated');
        push @ErrorMessages, { 'Message' => 'ValidCouponCountReached'}
            if (defined $CouponCampaign->get('ValidCouponCount')
                and $CouponCampaign->get('RedeemCount') >= $CouponCampaign->get('ValidCouponCount'));
        push @ErrorMessages, { 'Message' => 'CouponNotEffective'} unless $CouponCampaign->get('IsEffective') ;
    }

    my $aConditions    = $CouponCampaign->get('CouponConditions');
    my @PaymentMethodConditions  = map {$_->get('ConditionObject') } grep {$_->get('Type') eq 'PAYMENT'} @$aConditions;
    my @ShippingMethodConditions = map {$_->get('ConditionObject') } grep {$_->get('Type') eq 'SHIPPING'} @$aConditions;
    my @ProductConditions        = map {$_->get('ConditionObject') } grep {$_->get('Type') eq 'PRODUCT'} @$aConditions;

    if ( scalar(@PaymentMethodConditions) ) {
        my $PaymentLineItem = $LineItemCoupon->container->get('Payment');
        my $PaymentMethod  = defined $PaymentLineItem ? $PaymentLineItem->get('PaymentMethod'): undef;
        push @ErrorMessages, { 'Message' => 'PaymentMethodMisMatch'}
            unless defined $PaymentMethod and ElementOf($PaymentMethod, @PaymentMethodConditions);
    }
    if ( scalar(@ShippingMethodConditions) ) {
        my $ShippingLineItem = $LineItemCoupon->container->get('Shipping');
        my $ShippingMethod = defined $ShippingLineItem ? $ShippingLineItem->get('ShippingMethod') : undef;
        push @ErrorMessages, { 'Message' => 'ShippingMethodMisMatch'}
            unless defined $ShippingMethod and ElementOf($ShippingMethod, @ShippingMethodConditions);
    }
    if ( scalar(@ProductConditions) ) {
        my $Products = $LineItemCoupon->container->get('LineItems');
        my $ProductMisMatch = 1;
        foreach my $LineItem (@$Products) {
            if ($LineItem->instanceOf('LineItemProduct')) {
                my $Product = $LineItem->get('Product');
                next unless defined $Product;
                my $SuperProduct = $Product->get('SuperProduct');
                $ProductMisMatch = 0 if ( ElementOf($Product, @ProductConditions)
                                          or (defined $SuperProduct and ElementOf($SuperProduct, @ProductConditions)));
            }
        }
        push (@ErrorMessages , { 'Message' => 'ProductMisMatch'}) if $ProductMisMatch;
    }
    return \@ErrorMessages;
}

1;
