#========================================================================================
# package      DE_EPAGES::Saferpay2::API::Object::LineItemPaymentSaferpay2
# state        public
# base         DE_EPAGES::ExternalPayment::API::Object::LineItemPaymentExternal
#----------------------------------------------------------------------------------------
# description  object interface for LineItemPaymentSaferpay2.
#========================================================================================
package DE_EPAGES::Saferpay2::API::Object::LineItemPaymentSaferpay2;
use base DE_EPAGES::ExternalPayment::API::Object::LineItemPaymentExternal;

use DE_EPAGES::Core::API::Error qw ( Error );
use DE_EPAGES::Core::API::PerlTools qw ( GetCurrentDateTime );
use DE_EPAGES::Trigger::API::Trigger qw ( TriggerHook );
use DE_EPAGES::ExternalPayment::API::Log qw ( LogPayment );

use DE_EPAGES::Saferpay2::API::Constants qw (
    PAYMENT_TYPE_SAFERPAY2
);

use strict;


#========================================================================================
# function     className
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ClassName = $LineItemPaymentSaferpay2->className;
#----------------------------------------------------------------------------------------
# description  Returns the class name, used for DAL access.
#----------------------------------------------------------------------------------------
# return       $ClassName | class name | String
#========================================================================================
sub className {
    return 'LineItemPaymentSaferpay2';
}



#========================================================================================
# function     insert
# state        private
#----------------------------------------------------------------------------------------
# syntax       $LineItemPaymentSaferpay2 = DE_EPAGES::Saferpay2::API::Object::LineItemPaymentSaferpay2->insert($hInfo);
# example      my $Class = LoadClassByAlias('LineItemPaymentSaferpay2');
#               my $LineItemPaymentSaferpay2 = $Class->insertObject($hInfo);
#----------------------------------------------------------------------------------------
# description  Insert this info as object, afterwards the hook 'OBJ_InsertLineItemPaymentSaferpay2' is triggered.
#----------------------------------------------------------------------------------------
# input        $Package | class package | string
# input        $hInfo | attributes of new object | ref.hash
# return       $LineItemPaymentSaferpay2 | new object | object
# hook         OBJ_InsertLineItemPaymentSaferpay2 | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>LineItemPaymentSaferpay2ID - line item payment id - int
#               </ul>
#========================================================================================
sub insert {
    my ($Package, $hInfo) = @_;

    my $self = $Package->SUPER::insert($hInfo);

    TriggerHook('OBJ_InsertLineItemPaymentSaferpay2', {
        'Object' => $self,
        'LineItemPaymentSaferpay2ID' => $self->id(),
    });

    return $self;
}



#========================================================================================
# function     delete
# state        private
#----------------------------------------------------------------------------------------
# syntax       $LineItemPaymentSaferpay2->delete;
#----------------------------------------------------------------------------------------
# description  Deletes this object, therefore the hook 'OBJ_DeleteLineItemPaymentSaferpay2' is triggered.
#----------------------------------------------------------------------------------------
# hook         OBJ_DeleteLineItemPaymentSaferpay2 | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>LineItemPaymentSaferpay2ID - line item payment id - int
#               </ul>
#========================================================================================
sub delete {
    my $self = shift;

    TriggerHook('OBJ_DeleteLineItemPaymentSaferpay2', {
        'Object' => $self,
        'LineItemPaymentSaferpay2ID' => $self->id(),
    });

    $self->SUPER::delete;
    return;
}



#========================================================================================
# function     set
# state        private
#----------------------------------------------------------------------------------------
# syntax       $LineItemPaymentSaferpay2->set($hValues, $LanguageID);
#----------------------------------------------------------------------------------------
# description  This function triggers hooks 'OBJ_BeforeUpdateLineItemPaymentSaferpay2'
#               and 'OBJ_AfterUpdateLineItemPaymentSaferpay2'.
#----------------------------------------------------------------------------------------
# input        $hValues | attribute name => value hash | ref.hash
# input        $LanguageID | (optional; required for localized attributes) language id | int
#----------------------------------------------------------------------------------------
# hook         OBJ_BeforeUpdateLineItemPaymentSaferpay2 | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>LineItemPaymentSaferpay2ID - line item payment id - int
#                 <li>Values        - new values for object - ref.hash
#                 <li>LanguageID    - values of this language will be updated - int
#               </ul>
# hook         OBJ_AfterUpdateLineItemPaymentSaferpay2 | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>LineItemPaymentSaferpay2ID - line item payment id - int
#                 <li>Values        - updated values       - ref.hash
#                 <li>LanguageID    - values updated of this language - int
#               </ul>
#========================================================================================
sub set {
    my $self = shift;
    my ($hValues, $LanguageID) = @_;

    TriggerHook('OBJ_BeforeUpdateLineItemPaymentSaferpay2', {
        'Object'=>$self,
        'Values'=> $hValues,
        'LineItemPaymentSaferpay2ID' => $self->id(),
        'LanguageID' => $LanguageID,
    });

    $self->SUPER::set($hValues, $LanguageID);

    TriggerHook('OBJ_AfterUpdateLineItemPaymentSaferpay2', {
        'Object'=>$self,
        'Values'=> $hValues,
        'LineItemPaymentSaferpay2ID' => $self->id(),
        'LanguageID' => $LanguageID,
    });

    return;
}

#========================================================================================
# function     executePayment
# state        private
#----------------------------------------------------------------------------------------
# syntax       $LineItemPaymentSaferpay2->executePayment($hParams);
#----------------------------------------------------------------------------------------
# description  This function directs the payment process.
#----------------------------------------------------------------------------------------
# input        $hValues | environment information
#                       <ul>
#                           <li>Basket - basket object (optional)</li>
#                           <li>Servlet - servlet handling the current request</li>
#                           <li>Order - order object (optional)</li>
#                       </ul>
#                   | ref.hash
#========================================================================================
sub executePayment {
    my $self = shift;
    my ($hParams) = @_;

    $self->updateOnExecute($hParams);

    my $ViewAction =
      $hParams->{'Servlet'}->vars('IsMobile') ? 'MobileSF-ViewPaymentSaferpay2' : 'ViewPaymentSaferpay2';

    # show payment template
    my $Object = $hParams->{'Order'} // $hParams->{'Basket'};
    Error(
        'INVALID_FORM',
        {
            'ViewVars' => {
                'FormError'    => 0,
                'ViewObjectID' => $Object->id,
            },
            'ViewAction' => $ViewAction
        }
    );

    return;
}


sub updateTransactionStatus {
    my $self = shift;
    my ($TransID, $TransStatus, $ExternalStatus) = @_;

    my $hData = {
        'TransStatus'  => $TransStatus,
        'ExternalStatus'  => $ExternalStatus,
        'TransID' => $TransID,
        'TransTime'         => GetCurrentDateTime(),
    };
    $self->set($hData);
    LogPayment(PAYMENT_TYPE_SAFERPAY2, 'updateTransactionStatus', $hData);
}

1;