#========================================================================================
# package      DE_EPAGES::Saferpay2::API::Object::PaymentMethodSaferpay2
# state        public
# base         DE_EPAGES::Order::API::Object::ContainerPaymentMethod
#----------------------------------------------------------------------------------------
# description  object interface for Payment.
#========================================================================================
package DE_EPAGES::Saferpay2::API::Object::PaymentMethodSaferpay2;
use base DE_EPAGES::Order::API::Object::ContainerPaymentMethod;


use DE_EPAGES::Core::API::PerlTools qw ( isValid );


use DE_EPAGES::Saferpay2::API::Constants qw (
    INTERNAL_CARTRIDGE_NAME
    SUPPORTED_CURRENCIES
);

use strict;


#========================================================================================
# function     featureName
# state        public
#----------------------------------------------------------------------------------------
# syntax       $FeatureName = $PaymentMethod->featureName;
#----------------------------------------------------------------------------------------
# description   Returns the feature name 'Saferpay2'.
#----------------------------------------------------------------------------------------
# return       $FeatureName | feature name (undef means no feature count) | String
#========================================================================================
sub featureName {
     return INTERNAL_CARTRIDGE_NAME;
}


#========================================================================================
# function     usableAtStorefront
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Usable = $PaymentMethod->usableAtStorefront($hVars);
#----------------------------------------------------------------------------------------
# description  Tests if a payment method is visible in the storefront.
#               Overrides function <function DE_EPAGES::Payment::API::Object::PaymentMethod::usableAtStorefront>
#               to evaluate if payment method is configured
#----------------------------------------------------------------------------------------
# input        $hVars | <ul>
#                   <li>CurrencyID    - alpha currency code (ISO 4217) - char(3)
#                   <li>TaxModel      - gross or net price    - int
#               </ul> | ref.hash
# return       $Usable | true if payment method is usable in the storefront | boolean
#========================================================================================
sub usableAtStorefront {
    my $self = shift;
    my ($hVars) = @_;

    my $hInfo = $self->get([qw( MerchantName MerchantPassword CustomerId TerminalId)]);

    my $IsUsableAtStorefront = 0;
    if( $hInfo->{'MerchantName'} && $hInfo->{'MerchantPassword'} && $hInfo->{'CustomerId'} && $hInfo->{'TerminalId'} && $self->SUPER::usableAtStorefront($hVars) ) {
        $IsUsableAtStorefront = 1;
    }

    return $IsUsableAtStorefront;
}


#========================================================================================
# function     canAddToBasket
# state        public
#----------------------------------------------------------------------------------------
# syntax       if ($PaymentMethod->canAddToBasket($hVars)) {...
#----------------------------------------------------------------------------------------
# description  Tests if a payment method can be added to the basket with the given currency.
#               Overrides function <function DE_EPAGES::Order::API::Object::ContainerPaymentMethod::canAddToBasket>
#               Valid currencies are: EUR, CHF
#----------------------------------------------------------------------------------------
# input        $hVars | <ul>
#                   <li>CurrencyID    - alpha currency code (ISO 4217) - char(3)
#                   <li>TaxModel      - gross or net price    - int
#               </ul> | ref.hash
# return       $canAddToBasket | true if payment method can be used for the given currency | boolean
#========================================================================================
sub canAddToBasket {
    my $self = shift;
    my ($hVars) = @_;

    my $AreCurrenciesSupported = $hVars->{'CurrencyID'} ~~ [(SUPPORTED_CURRENCIES)];
    my $CanBeAddedToBasket = $AreCurrenciesSupported && $self->SUPER::canAddToBasket($hVars);

    return $CanBeAddedToBasket;
}


sub insert {
    my $class = shift;
    my ($hInfo) = @_;

    my $self = $class->SUPER::insert($hInfo);
    $self->_setDefaultNameForAllLanguages();
    return $self;
}


sub default {
    my $class = shift;
    my ($Parent, $hInfo) = @_;

    # delete preset name so we can set in later for all languages.
    delete($hInfo->{'Name'});

    return {
        %{$class->SUPER::default($Parent, $hInfo)},
        'IsInTestMode' => 0,
        'DisplayLogo' => 0
    };
}


sub _setDefaultNameForAllLanguages {
    my $self = shift;

    my $PaymentTypeSaferpay2 = $self->get('PaymentType');
    my $Shop = $self->get('Shop');
    my $PaymentMethods = $Shop->folder('PaymentMethods');
    my $hInstalledLanguages = $Shop->get('Languages');
    my $Alias;

    foreach my $Language (@{ $hInstalledLanguages }) {

        my $LanguageID = $Language->{'LanguageID'};
        my $LanguageDefaultName = $PaymentTypeSaferpay2->get('PaymentMethodDefaultName', $LanguageID);

        if(isValid($LanguageDefaultName)) {
            $self->set({'Name' => $LanguageDefaultName}, $LanguageID);
            $Alias = $self->_getUniqueAliasForPayment($PaymentMethods, $LanguageDefaultName);
            $self->set({'Alias' => $Alias}, $LanguageID);
        }
    }
}


sub _getUniqueAliasForPayment{
    my $self = shift;
    my ($PaymentMethods, $LanguageDefaultName) = @_;

    if ($PaymentMethods->existsChild($LanguageDefaultName) and $self->get('Alias') ne $LanguageDefaultName) {
        my $i = 1;
        my $newAliasMethod = $LanguageDefaultName . '_' .  $i;
        while ($PaymentMethods->existsChild($newAliasMethod) and $self->get('Alias') ne $newAliasMethod) {
            $newAliasMethod =  $LanguageDefaultName . '_' .  ++$i;
        }
        $LanguageDefaultName = $newAliasMethod;
    }
    return $LanguageDefaultName;
}


1;