#========================================================================================
# package      DE_EPAGES::Saferpay2::API::SaferpayService
# state        ublic
#----------------------------------------------------------------------------------------
# Saferpay Service
#========================================================================================

package DE_EPAGES::Saferpay2::API::SaferpayService;

use DE_EPAGES::ExternalPayment::API::Log qw ( LogPayment );
use DE_EPAGES::Saferpay2::API::Constants qw (
    PAYMENT_TYPE_SAFERPAY2
    SUPPORTED_PAYMENT_METHODS
);
use DE_EPAGES::Object::API::Factory qw ( LoadRootObject );
use DE_EPAGES::Object::API::Table::Currency qw ( GetInfoCurrency );
use DE_EPAGES::Core::API::PerlTools qw (GenerateGUID);

use HTTP::Request;
use HTTP::Status qw ( is_server_error );
use JSON::MaybeXS qw ( encode_json decode_json );
use LWP::UserAgent;

use strict;


#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       DE_EPAGES::Saferpay2::API::SaferpayService->new( %args );
#----------------------------------------------------------------------------------------
# description  Instantiates a new SaferpayService.
#----------------------------------------------------------------------------------------
# input        $RequestObject | request object to handle | object
# return       SaferpayService | object | DE_EPAGES::Saferpay2::API::SaferpayService
#========================================================================================
sub new {
    my $class = shift;
    my ( $Order ) = @_;

    my $LineItemContainer = $Order->get('LineItemContainer');
    my $LineItemPayment = $LineItemContainer->get('Payment');
    my $PaymentMethod = $LineItemPayment->get('PaymentMethod');

    my $self  = bless {
        Order             => $Order,
        LineItemContainer => $LineItemContainer,
        LineItemPayment => $LineItemPayment,
        PaymentMethod     => $PaymentMethod
    }, $class;

    return $self;
}


#========================================================================================
# function     InitializePaymentPage
# state        public
#----------------------------------------------------------------------------------------
# syntax       DE_EPAGES::Saferpay2::API::SaferpayService->InitializePaymentPage( $PaymentWebhookUrl,
# $ShopperRedirectUrl );
#----------------------------------------------------------------------------------------
# description  With this request, you forward all data necessary for the payment to Saferpay.
#----------------------------------------------------------------------------------------
# input        $PaymentWebhookUrl  | NotifyUrl to notify the merchant system, as a redundant measure  | string
# input        $ShopperRedirectUrl | request object to handle | string
# return       Response of PaymentPage Initialize | hashref
#========================================================================================
sub InitializePaymentPage {
    my $self    = shift;
    my ($PaymentWebhookUrl, $ShopperRedirectUrl) = @_;

    my $System = LoadRootObject();
    my $ApiBaseUrl = $System->get('Saferpay2_WebserviceEndpointURL');


    my $Shop = $self->{'Order'}->getSite();
    my $Url = $ApiBaseUrl."Payment/v1/PaymentPage/Initialize";
    my $Header = ['Content-Type' => 'application/json; charset=UTF-8'];
    my $hContainerInfo = $self->{'LineItemContainer'}->get([ qw( CurrencyID GrandTotal )]);

    my $hData = {
        RequestHeader => {
            SpecVersion    => "1.18",
            CustomerId     => $self->{'PaymentMethod'}->get('CustomerId'),
            RequestId      => GenerateGUID(),
            RetryIndicator => 0
        },
        TerminalId    => $self->{'PaymentMethod'}->get('TerminalId'),
        Payment       => {
            Amount      => {
                Value        => $self->_FormatFloatForSaferpay($hContainerInfo->{'GrandTotal'}, $hContainerInfo->{'CurrencyID'}),
                CurrencyCode => $hContainerInfo->{'CurrencyID'}
            },
            OrderId     => $self->{'Order'}->alias(),
            Description => $Shop->get('NameOrAlias')
        },
        PaymentMethods => SUPPORTED_PAYMENT_METHODS,
        ReturnUrls    => {
            Success => $ShopperRedirectUrl,
            Fail    => $ShopperRedirectUrl
        },
        Notification  => {
            NotifyUrl => $PaymentWebhookUrl
        }
    };
    $self->_LogPayment('InitializePaymentPage Request', {'request' => $hData});
    my $hResponse = $self->_doPostRequest($Url, $Header, encode_json($hData));
    $self->_LogPayment('InitializePaymentPage Response', {'response' => $hResponse});

    return $hResponse;
}


sub DoAssertCall {
    my $self = shift;

    my $System = LoadRootObject();
    my $ApiBaseUrl = $System->get('Saferpay2_WebserviceEndpointURL');
    my $Url = $ApiBaseUrl . "Payment/v1/PaymentPage/Assert";
    my $Header = [ 'Content-Type' => 'application/json; charset=UTF-8' ];

    $self->_LogPayment('ORDER: ', { 'order' => $self->{'Order'} });

    my $hData = {
        RequestHeader => {
            SpecVersion    => "1.18",
            CustomerId     => $self->{'PaymentMethod'}->get('CustomerId'),
            RequestId      => GenerateGUID(),
            RetryIndicator => 0
        },
        Token         => $self->{'LineItemPayment'}->get('TransID2'),
    };

    $self->_LogPayment('DoAssertCall Request', { 'request' => $hData });
    my $hResponse = $self->_doPostRequest($Url, $Header, encode_json($hData));
    $self->_LogPayment('DoAssertCall Response', { 'response' => $hResponse });
    return $hResponse;
}


sub _doPostRequest {
    my $self = shift;
    my ($Url, $Header, $Data) = @_;

    my $request = HTTP::Request->new('POST', $Url, $Header, $Data);
    $request->authorization_basic(
        $self->{'PaymentMethod'}->get('MerchantName'), $self->{'PaymentMethod'}->get('MerchantPassword')
    );
    my $userAgent = LWP::UserAgent->new();

    my $res = $userAgent->request($request);
    my $hResponse;
    if ( is_server_error($res->code) ) {
        $hResponse = {
            Behavior     => 'ABORT',
            ErrorMessage => 'Internal Server Error',
            ErrorName => 'INTERNAL_SERVER_ERROR'
        }
    } else {
        $hResponse = decode_json($res->decoded_content());
    }
    return $hResponse;
}

sub _LogPayment {
    my $self = shift;
    my ($Name, $DebugData, $Section, $Filter) = @_;

    return LogPayment(PAYMENT_TYPE_SAFERPAY2, $Name, $DebugData, $Section, $Filter);
}

#========================================================================================
# function     FormatFloatForSaferpay
# state        public
#----------------------------------------------------------------------------------------
# syntax       $FormattedFloatNumber = FormatFloatForSaferpay($FloatNumber);
#----------------------------------------------------------------------------------------
# description  It formats float number according to Saferpay
#----------------------------------------------------------------------------------------
# input        $FloatNumber | float number | float
# return       $FormattedFloatNumber | formatted float number string | string
#========================================================================================
sub _FormatFloatForSaferpay {
    my $self = shift;
    my ($amount, $currencyid) = @_;

    my $hCurrency = GetInfoCurrency($currencyid);
    my $price = sprintf('%.0f', $amount * 10 ** $hCurrency->{'MinorUnit'});
    return $price;
}

1;