#========================================================================================
# package      DE_EPAGES::Saferpay2::UI::Shop
# base         DE_EPAGES::Order::UI::Shop
# state        public
#----------------------------------------------------------------------------------------
# description  UI Controller for shop
#========================================================================================
package DE_EPAGES::Saferpay2::UI::Shop;
use base DE_EPAGES::Order::UI::Shop;
use DE_EPAGES::Core::API::Error qw ( Error ExistsError GetError );
use DE_EPAGES::ExternalPayment::API::Log qw ( LogPayment );
use DE_EPAGES::Saferpay2::API::Constants qw (
    INTERNAL_CARTRIDGE_NAME
    PAYMENT_TYPE_SAFERPAY2
);
use DE_EPAGES::Saferpay2::API::Tools qw ( HandleTransactionResponse );
use DE_EPAGES::Saferpay2::API::SaferpayService;


use strict;


#========================================================================================
# function     TransactionPaymentName
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Name = $Handler->TransactionPaymentName();
#----------------------------------------------------------------------------------------
# description  Name of PaymentType for search/view transactions.
#----------------------------------------------------------------------------------------
# return       $Name | name of payment type | string
#========================================================================================
sub TransactionPaymentName {
    return INTERNAL_CARTRIDGE_NAME;
}


#========================================================================================
# function     PaymentBacklinkSaferpay2
# state        public
#----------------------------------------------------------------------------------------
# syntax       PaymentBacklinkSaferpay2( $Servlet );
#----------------------------------------------------------------------------------------
# description  Redirect URL for the shopper in case of success or error
#----------------------------------------------------------------------------------------
# input        $Servlet | Servlet | DE_EPAGES::Presentation::Servlet::BasePermission
#========================================================================================
sub PaymentBacklinkSaferpay2 {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop = $Servlet->object();

    my $hQueryParams = $Servlet->params();
    my $OrderId = $hQueryParams->{'OrderId'};
    my $LineItemPaymentGUID = $hQueryParams->{'LineItemPaymentGUID'};

    my $Order = $Shop->loadObject($OrderId, 'CustomerOrder');

    $self->_LogPayment('PaymentBacklinkSaferpay2', {'hQueryParams' => $hQueryParams});

    if ( $self->_isPaymentGUIDCorrect($Order, $LineItemPaymentGUID) ) {
        $Servlet->vars('ViewAction', 'View');
        $Servlet->vars('ViewObjectID', $OrderId);
    };

    my $SaferpayService = DE_EPAGES::Saferpay2::API::SaferpayService->new($Order);
    my $hResponse = $SaferpayService->DoAssertCall();

    my $LineItemPayment = $Order->get('LineItemContainer')->get('Payment');

    $hResponse = HandleTransactionResponse($hResponse);

    if ( $hResponse->{'HasErrorsInResponse'} ) {
        $Servlet->vars('ViewAction', 'View');
        $Servlet->vars('ViewObjectID', $Order->id());
        $Servlet->vars('SaferpayPaymentFailed', 1);
        $Servlet->vars('SaferpayPaymentFailedReason', $hResponse->{'ErrorName'});
    }

    $LineItemPayment->updateTransactionStatus(
        $hResponse->{'Transaction'}->{'Id'}, $hResponse->{'Transaction'}->{'Status'}, $hResponse->{'Transaction'}->{'ExternalStatus'}
    );

    return;
}


#========================================================================================
# function     PaymentCallbackSaferpay2
# state        public
#----------------------------------------------------------------------------------------
# syntax       PaymentCallbackSaferpay2( $Servlet );
#----------------------------------------------------------------------------------------
# description  Server-server callback from payment provider. Webhook for status change to
#               success
#----------------------------------------------------------------------------------------
# input        $Servlet | Servlet | DE_EPAGES::Presentation::Servlet::BasePermission
#========================================================================================
sub PaymentCallbackSaferpay2 {
    my $self = shift;
    my ($Servlet) = @_;

    my $hQueryParams = $Servlet->params();
    my $OrderId = $hQueryParams->{'OrderId'};
    my $LineItemPaymentGUID = $hQueryParams->{'LineItemPaymentGUID'};

    my $Shop;
    my $Order;
    eval {
        $Shop = $Servlet->object();
        $Order = $Shop->loadObject($OrderId, 'CustomerOrder');
        my $LineItemPayment = $Order->get('LineItemContainer')->get('Payment');

        $self->_LogPayment('PaymentCallbackSaferpay2', {ServletParams => $Servlet->params()});
        if ( !$self->_isPaymentGUIDCorrect($Order, $LineItemPaymentGUID) ) {
            Error('LineItemPaymentGUID_Order_MISMATCH');
        }
        my $SaferpayService = DE_EPAGES::Saferpay2::API::SaferpayService->new($Order);
        my $hResponse = $SaferpayService->DoAssertCall();

        $hResponse = HandleTransactionResponse($hResponse);
        $LineItemPayment->updateTransactionStatus(
            $hResponse->{'Transaction'}->{'Id'}, $hResponse->{'Transaction'}->{'Status'}, $hResponse->{'Transaction'}->{'ExternalStatus'}
        );


    };
    if (ExistsError()) {
        my $Error = GetError();
        $self->_LogPayment('PaymentCallbackSaferpay2 failed', {
            'Error'               => $Error->message(),
            'OrderId'             => $OrderId,
            'LineItemPaymentGUID' => $LineItemPaymentGUID
        });
    }

    $Servlet->params('ContentType', 'text/plain');
    $Servlet->params('ExitAfterEvent', 1);
    $Servlet->writeContentOutput('OK');
    return;
}

sub _LogPayment {
    my $self = shift;
    my ($Name, $DebugData, $Section, $Filter) = @_;

    return LogPayment(PAYMENT_TYPE_SAFERPAY2, $Name, $DebugData, $Section, $Filter);
}

sub _isPaymentGUIDCorrect {
    my $self = shift;
    my ($Order, $LineItemPaymentGUID) = @_;

    return ($Order->get('LineItemContainer')->get('Payment')->guid() eq $LineItemPaymentGUID);
}

1;