#========================================================================================
# package      DE_EPAGES::Spreedly::UI::BasketOrder
# base         DE_EPAGES::Presentation::UI::Object
# state        private
#----------------------------------------------------------------------------------------
# description  UI Controller for basket or order
#========================================================================================
package DE_EPAGES::Spreedly::UI::BasketOrder;
use base DE_EPAGES::Presentation::UI::Object;

use DE_EPAGES::Core::API::Error qw ( ExistsError GetError );
use DE_EPAGES::Core::API::PerlTools qw ( isValid );
use DE_EPAGES::ExternalPayment::API::Log qw ( LogPayment );
use DE_EPAGES::Order::API::Object::BasketOrderProcess;
use DE_EPAGES::Order::UI::Basket;
use DE_EPAGES::Spreedly::API::PaymentInformationController;
use DE_EPAGES::Spreedly::API::PurchaseController;
use DE_EPAGES::Spreedly::API::Constants qw ( 
    PURCHASE_RESPONSE_STATE_SUCEEDED
    PURCHASE_RESPONSE_STATE_PROCESSING
    FAILURE_MESSAGES
    PAYMENT_STATUS_PAID
    PAYMENT_STATUS_FAILED );
use JSON qw ( decode_json );

use feature 'switch';

use strict;

#========================================================================================
# function     ConfirmSuccessfulTransaction
# state        public
#----------------------------------------------------------------------------------------
# syntax       ConfirmSuccessfulTransaction( $Servlet, $BaketOrder );
#----------------------------------------------------------------------------------------
# description  Retrieves Payment Status Information.
#----------------------------------------------------------------------------------------
# input        $Servlet | Servlet | object
# input        $BasketOrder | Basket or Order object | object
#========================================================================================
sub ConfirmSuccessfulTransaction {
    my $self = shift;
    my ($Servlet, $BasketOrder) = @_;

    my $LineItemContainer = $BasketOrder->get('LineItemContainer');
    my $LineItemPaymentSpreedly = $LineItemContainer->get('Payment');
    my $TransactionToken = $LineItemPaymentSpreedly->getTransactionToken();

    if (isValid($Servlet->params()->{'ErrorName'}) && isValid($Servlet->params()->{'ErrorReason'})) {
        $self->_setSpreedlyPaymentError($Servlet, $Servlet->params()->{'ErrorName'}, $Servlet->params()->{'ErrorReason'});
        return;
    }

    if (not $LineItemPaymentSpreedly->isPaymentSuccessful()) {

        my $hTransactionResult = $self->checkTransactionStateOnSpreedly($Servlet, $BasketOrder, $TransactionToken);
        if(defined $hTransactionResult->{'Transaction'}) {
            my $PaymentStatus = $LineItemPaymentSpreedly->getPaymentStatusFromTransactionState($hTransactionResult->{'Transaction'}->{'state'});

            $LineItemPaymentSpreedly->updateTransactionStatus($PaymentStatus, $hTransactionResult->{'Transaction'});
            $self->_LogPayment($LineItemPaymentSpreedly, 'Update transaction status Succeeded On Spreedly', $hTransactionResult);

            if ($LineItemPaymentSpreedly->isPaymentProcessing()) {
                $self->_handleProcessingPayment($Servlet, $BasketOrder, $LineItemPaymentSpreedly);
            }

            return;
        }

        $self->_ShowPaymentFailure($Servlet, FAILURE_MESSAGES->{'Name'}->{'Purchase'}, FAILURE_MESSAGES->{'Reason'}->{'UnknownError'});
        return;
    }

    return;
}

sub checkTransactionStateOnSpreedly {
    my $self = shift;
    my ($Servlet, $BasketOrder, $TransactionToken) = @_;
    my $hTransaction;

    eval {
        $hTransaction = $self->_showTransaction($BasketOrder, $TransactionToken);
    };
    if ( ExistsError() ) {
        my $Error = GetError();
        return {
            'Transaction' => undef,
            'ErrorMessage' => $Error->message,
        };
    }
    if ( $hTransaction->{'successful'} &&
        ( $hTransaction->{'state'} eq PURCHASE_RESPONSE_STATE_SUCEEDED ||
        $hTransaction->{'state'} eq PURCHASE_RESPONSE_STATE_PROCESSING )
        ) {
        return {
            'Transaction'  => $hTransaction,
            'ErrorMessage' => undef
        };
    }
    return {
            'Transaction' => undef,
            'ErrorMessage' => 'UnknownError',
        };
}

#----------------------------------------------------------------------------------------
sub _showTransaction {
    my $self = shift;
    my ($BasketOrder, $TransactionToken) = @_;

    my $Shop        = $BasketOrder->getSite();
    my $GatewayName = $self->_getGatewayName($BasketOrder);
    my $PurchaseController = DE_EPAGES::Spreedly::API::PurchaseController->new();
    my $hTransaction = $PurchaseController->ShowTransaction($Shop, $GatewayName, $TransactionToken);

    return $hTransaction;
}


#========================================================================================
# function     RetrievePaymentInfoToken
# state        public
#----------------------------------------------------------------------------------------
# syntax       RetrievePaymentInfoToken( $Servlet );
#----------------------------------------------------------------------------------------
# description  Retrieves Payment Information Token.
#----------------------------------------------------------------------------------------
# input        $Servlet | Servlet | object
#========================================================================================
sub RetrievePaymentInfoToken {
    my $self = shift;
    my ($Servlet) = @_;

    my $Basket = $Servlet->object();
    my $LineItemPayment = $self->_RetrieveLineItemPayment($Basket);
    my $PaymentInfoToken = $Servlet->params()->{'payment_method_token'};

    $self->_LogPayment($LineItemPayment, 'Payment Information Token', $PaymentInfoToken);

    my $Form = $Servlet->form();
    if (not isValid($PaymentInfoToken)) {
        $Form->executeFormError({
            'Name'       => 'SpreedlyPaymentInfoTokenResult',
            'Reason'     => 'ConnectionError',
            'ViewAction' => $Form->value('ViewActionCurrent'),
        });
    }
    # GatewayFieldBirthdate is required for payu_latam
    my $GatewayFieldBirthday = $Servlet->params()->{'GatewayFieldBirthdate'};
    if(isValid($GatewayFieldBirthday)) {
        $Servlet->vars('GatewayFieldBirthday', $GatewayFieldBirthday);
    }

    $self->_StoreTokenFromController($LineItemPayment, $PaymentInfoToken);

    return;
}

#========================================================================================
# function     MobileMakePurchase
# state        public
#----------------------------------------------------------------------------------------
# syntax       MobileMakePurchase( $Servlet );
#----------------------------------------------------------------------------------------
# description  ChangeAction for mobileSF that calls makePurchase
#----------------------------------------------------------------------------------------
# input        $Servlet | Servlet | DE_EPAGES::Presentation::Servlet::BasePermission
#========================================================================================
sub MobileMakePurchase {
    my $self = shift;
    my ($Servlet) = @_;

    my $Basket = $Servlet->object();
    my $hParams = {
        'Basket' => $Basket,
        'Servlet' => $Servlet,
    };
    my $LineItemContainer = $Basket->get('LineItemContainer');
    my $LineItemPaymentSpreedly = $LineItemContainer->get('Payment');
    $LineItemPaymentSpreedly->updateOnExecute($hParams);

    $self->_makePurchase($Servlet, $Basket, $LineItemContainer);

    return;
}

#========================================================================================
# function     _makePurchase
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_makePurchase( $Servlet, $Basket, $LineItemContainer );
#----------------------------------------------------------------------------------------
# description  UI Controller for purchase at multi-step checkout
#----------------------------------------------------------------------------------------
# input        $Servlet | Servlet | DE_EPAGES::Presentation::Servlet::BasePermission
# input        $Basket | Basket | Object
# input        $LineItemContainer | LineItemContainer | Object
#========================================================================================

sub _makePurchase {
    my $self = shift;
    my ($Servlet, $Basket, $LineItemContainer) = @_;

    my $SpreedlyBrowserInfo = $Servlet->params()->{'spreedly_browser_info'};

    my $hTransaction;

    eval { $hTransaction = $self->_makePurchaseRequest( $LineItemContainer, $Basket, $SpreedlyBrowserInfo); };
    if ( ExistsError() ) {
        $self->_HandleCriticalErrorOnPurchaseRequest($Basket, $Servlet);
    }

    my $LineItemPayment = $self->_RetrieveLineItemPayment($Basket);
    $self->_LogPayment($LineItemPayment, 'Make Purchase', { 'Basket' =>  $Basket->guid(), 'Transaction' => $hTransaction });
    my $Order = $self->_createCustomerOrder( $Servlet, $Basket );

    if ( $hTransaction->{'successful'} && $hTransaction->{'state'} eq PURCHASE_RESPONSE_STATE_SUCEEDED) {
        $self->_SaveTransactionNumber( $hTransaction, $Order );
        $self->_SetOrderToPaid($Order);
        $self->_ShowCustomerOrder($Order, $Servlet);
    } else {
        $self->_LogPayment($LineItemPayment, 'Purchase Response or transaction state not sucessful', $hTransaction);

        $self->_SetPaymentAsFailed($Order);

        my $Reason = $hTransaction->{'successful'} ? FAILURE_MESSAGES->{'Reason'}->{'ConnectionError'} : FAILURE_MESSAGES->{'Reason'}->{'CardDeclined'};
        $self->_ShowPaymentFailure($Servlet, FAILURE_MESSAGES->{'Name'}->{'MobilePurchase'}, $Reason);
    }
    return;
}
#----------------------------------------------------------------------------------------
sub _setSpreedlyPaymentError {
    my $self = shift;
    my ($Servlet, $ErrorName, $ErrorReason) = @_;

    my $LineItemPayment = $self->_RetrieveLineItemPayment($Servlet->object());
    $self->_LogPayment($LineItemPayment, 'Set Spreedly Payment Error',
        {
            'ErrorName' => $ErrorName,
            'ErrorReason' => $ErrorReason,
        });

    my $ErrorDescription = {
        'Name' => FAILURE_MESSAGES->{'Name'}->{'Purchase'},
        'Reason' => FAILURE_MESSAGES->{'Reason'}->{'UnknownError'},
    };

    if (defined(FAILURE_MESSAGES->{'Name'}->{$ErrorName}) && defined(FAILURE_MESSAGES->{'Reason'}->{$ErrorReason})) {
        $ErrorDescription->{'Name'} = FAILURE_MESSAGES->{'Name'}->{$ErrorName};
        $ErrorDescription->{'Reason'} = FAILURE_MESSAGES->{'Reason'}->{$ErrorReason};
    }

    $self->_ShowPaymentFailure($Servlet, $ErrorDescription->{'Name'}, $ErrorDescription->{'Reason'});
    return;
}

sub _HandleCriticalErrorOnPurchaseRequest {
    my $self = shift;
    my ($Basket, $Servlet) = @_;

    my $Error = GetError();

    my $Response = $Error->vars->{Response};

    my $Reason = 'ConnectionError';
    my $ErrorMessage = '';

    # Not an empty string and json format
    if(!$Response =~ /^$/ && $Response =~ /^\{/) {
        my $hResponse = decode_json($Response);

        $Reason = $hResponse->{'code'};
        $ErrorMessage = $hResponse->{'message'};
    }

    my $LineItemPayment = $self->_RetrieveLineItemPayment($Basket);
    $self->_LogPayment($LineItemPayment, 'Make Purchase Request die ocurred',
        {
            'Basket' => $Basket->guid() ,
            'ErrorMessageEp6' => $Error->message,
            'ErrorMessageEpJ' => $ErrorMessage,
        }
    );

    $self->_ShowPaymentFailure($Servlet, FAILURE_MESSAGES->{'Name'}->{'MobilePurchase'}, $Reason);
    return;
}

#----------------------------------------------------------------------------------------
sub _ShowPaymentFailure {
    my $self = shift;
    my ($Servlet, $Name, $Reason) = @_;

    my $ViewAction = ($Servlet->vars('IsMobile')) ? 'View' : 'ViewMultiCheckoutPayment';

    my $LineItemPayment = $self->_RetrieveLineItemPayment($Servlet->object());
    $self->_LogPayment($LineItemPayment, 'Show Payment Failure', {
           'Name' => $Name,
           'Reason' => $Reason,
           'ViewAction' => $ViewAction
        });

    $Servlet->vars('ViewAction', $ViewAction);
    my $Form = $Servlet->form();
    $Form->executeFormError(
        {
           'Name' => $Name,
           'Reason' => $Reason
        }
    );
    return;
}

#----------------------------------------------------------------------------------------
sub _ShowCustomerOrder {
    my $self = shift;
    my ($Order, $Servlet) = @_;

    if (defined  $Order) {
        $Servlet->vars( 'ViewObjectID', $Order->id() );
    }

    return;
}

#----------------------------------------------------------------------------------------
sub _SaveTransactionNumber {
    my $self = shift;
    my ( $hTransaction, $Basket ) = @_;

    my $LineItemContainer       = $Basket->get('LineItemContainer');
    my $LineItemPaymentSpreedly = $LineItemContainer->get('Payment');
    $LineItemPaymentSpreedly->updateTransactionStatus(PAYMENT_STATUS_PAID, $hTransaction);

    return;
}

sub _SetOrderToPaid {
    my $self = shift;
    my ( $Order ) = @_;

    my $LineItemContainer = $Order->get('LineItemContainer');
    my $LineItemPaymentSpreedly = $LineItemContainer->get('Payment');

    $LineItemPaymentSpreedly->setOrderToPaid();

    return;
}

sub _SetPaymentAsFailed {
    my $self = shift;
    my ($Order) = @_;

    my $LineItemContainer       = $Order->get('LineItemContainer');
    my $LineItemPaymentSpreedly = $LineItemContainer->get('Payment');
    $LineItemPaymentSpreedly->updateTransactionStatus(PAYMENT_STATUS_FAILED);

    return;
}

sub _createCustomerOrder {
    my $self = shift;
    my ( $Servlet, $BasketOrder ) = @_;

    return $BasketOrder unless $BasketOrder->instanceOf('Basket');
    return DE_EPAGES::Order::UI::Basket->doBasket2Order( $Servlet, $BasketOrder );
}

#----------------------------------------------------------------------------------------
# Wrappers
#----------------------------------------------------------------------------------------

sub _getGatewayName {
    my $self = shift;
    my ( $Basket ) = @_;

    my $LineItemContainer       = $Basket->get('LineItemContainer');
    my $LineItemPaymentSpreedly = $LineItemContainer->get('Payment');
    my $PaymentMethod = $LineItemPaymentSpreedly->get('PaymentMethod');
    my $PaymentType = $PaymentMethod->get('PaymentType');

    return $PaymentType->get('GatewayName');
}

#----------------------------------------------------------------------------------------
sub _LogPayment {
    my $self = shift;
    my ($LineItemPayment, $Name, $DebugStuff) = @_;

    my $PaymentMethod   = $LineItemPayment->get('PaymentMethod');
    my $PaymentTypeAlias = $PaymentMethod->get('PaymentType')->alias();

    LogPayment($PaymentTypeAlias, $Name, $DebugStuff);

    return;
}

#----------------------------------------------------------------------------------------
sub _makePurchaseRequest {
    my $self = shift;
    my ( $LineItemContainer, $Basket, $SpreedlyBrowserInfo) = @_;

    my $Shop             = $Basket->getSite();
    my $GatewayName = $self->_getGatewayName($Basket);
    my $PurchaseController = DE_EPAGES::Spreedly::API::PurchaseController->new();

    return $PurchaseController->makePurchase( $LineItemContainer, $Shop, $GatewayName, $SpreedlyBrowserInfo);
}

#----------------------------------------------------------------------------------------
sub _StoreTokenFromController {
    my $self = shift;
    my ($LineItemPayment, $Token) = @_;

    my $Controller = DE_EPAGES::Spreedly::API::PaymentInformationController->new($LineItemPayment);
    $Controller->storePaymentInformation($Token);

    return;
}

#----------------------------------------------------------------------------------------
sub _RetrieveLineItemPayment {
    my $self = shift;
    my ($Basket) = @_;

    my $LineItemContainer = $Basket->container();

    return $LineItemContainer->get('Payment');
}

#----------------------------------------------------------------------------------------
sub _handleProcessingPayment {
    my $self = shift;
    my ($Servlet, $BasketOrOrder, $LineItemPaymentSpreedly) = @_;

    my $Shop = $BasketOrOrder->getSite();
    my $IsUnityRegistered = $Shop->get('IsUnityRegistered');

    $self->_LogPayment($LineItemPaymentSpreedly, 'Handle Processing Payment', { 'IsUnityRegistered' => $IsUnityRegistered });

    if ($IsUnityRegistered) {
        $LineItemPaymentSpreedly->set({'PaymentErrorCode' => 'Failed'});
        return;
    }

    $Servlet->vars('PAYMENT_STATUS_PROCESSING', 1);
    return;
}

1;
