#========================================================================================
# package      DE_EPAGES::Spreedly::UI::JSON::BasketOrder
# base         DE_EPAGES::Presentation::UI::JSON
# state        private
#----------------------------------------------------------------------------------------
# description  UI Controller for basket or order
#========================================================================================
package DE_EPAGES::Spreedly::UI::JSON;
use base DE_EPAGES::Presentation::UI::JSON;

use Time::HiRes  qw (gettimeofday);
use Digest::MD5 qw ( md5_hex );
use DE_EPAGES::Core::API::Error qw ( Error ExistsError GetError );
use DE_EPAGES::ExternalPayment::API::Log qw ( LogPayment );
use DE_EPAGES::Core::API::PerlTools qw ( isValid );
use DE_EPAGES::Shop::API::Url qw ( BuildShopUrl );
use DE_EPAGES::Presentation::API::DereferrerToken qw (GenerateDereferrerToken);
use DE_EPAGES::Order::API::Object::BasketOrderProcess;
use DE_EPAGES::Spreedly::API::PaymentInformationController;
use DE_EPAGES::Spreedly::API::PurchaseController;
use DE_EPAGES::Spreedly::API::Constants qw ( 
    PURCHASE_RESPONSE_REQUIRED_ACTION_REDIRECT
    PURCHASE_RESPONSE_STATE_SUCEEDED
    PURCHASE_RESPONSE_STATE_PENDING
    PAYMENT_STATUS_PAID
    JSON_FAILURE_MESSAGES
    THREE_DS_AVAILABLE_GATEWAYS
    SPREEDLY_GATEWAY_NAME_PAYU_LATAM );
use constant COUNTRY_CODE2_MX => 'MX';
use JSON qw ( decode_json );

use feature 'switch';

use strict;

#========================================================================================
# function     MakePurchase
# state        public
#----------------------------------------------------------------------------------------
# syntax       MakePurchase( $Servlet );
#----------------------------------------------------------------------------------------
# description  UI Controller for purchase at multi-step checkout
#----------------------------------------------------------------------------------------
# input        $Servlet | Servlet | DE_EPAGES::Presentation::Servlet::BasePermission
#========================================================================================
sub MakePurchase {
    my $self = shift;
    my ($Servlet) = @_;

    my $hReturnData = {
        'Error' => {
            'Name' => JSON_FAILURE_MESSAGES->{'Name'}->{'Purchase'},
            'Reason' => JSON_FAILURE_MESSAGES->{'Reason'}->{'UnknownError'},
        }
    };

    my $Basket = $Servlet->object();
    my $LineItemContainer = $Basket->get('LineItemContainer');
    my $hTransaction;

    my $PurchaseParams = {
        "SpreedlyBrowserInfo" => $Servlet->params()->{'SpreedlyBrowserInfo'},
        "RedirectUrl" => $self->_getRedirectUrl($Servlet, $Basket, $Servlet->params()),
        "CallbackUrl" => $self->_getCallbackUrl($Servlet, $Basket)
    };

    # gateway fields birthday and device session id for PayU_Latam
    eval {
        $self->_validateGatewayRequiredFields($PurchaseParams, $Basket, $Servlet);
    };
    if (ExistsError()) {
        my $Error = GetError();
        $hReturnData = {
            'Error' => {
                'Name' => JSON_FAILURE_MESSAGES->{'Name'}->{'Purchase'},
                'Reason' => $Error->message(),
            }
        };
        $self->_LogPayment($Servlet, 'Error on GatewayRequiredFields', {'Error' => $Error->message()});
        return $self->_writeDataToOutput($hReturnData);
    }

    $self->_LogPayment($Servlet, 'Make Purchase Request', $PurchaseParams);
    eval { $hTransaction = $self->_makePurchase( $LineItemContainer, $Basket, $PurchaseParams); };
    if ( ExistsError() ) {
        $hReturnData = $self->_HandleCriticalErrorOnRequest($Basket, $Servlet, 'Make Purchase Request die ocurred');
        return $self->_writeDataToOutput($hReturnData);
    }

    $self->_LogPayment($Servlet, 'Make Purchase Response Transaction and Basket GUID',
            { 'Basket' =>  $Basket->guid(), 'Transaction' => $hTransaction });

    if ( $hTransaction->{'successful'} ) {
        $self->_SaveTransactionToken($hTransaction->{'token'}, $LineItemContainer);
        $hReturnData = $self->_HandleSuccessfulPurchaseResponse($hTransaction, $Basket, $Servlet);
    } else {
        $self->_LogPayment($Servlet, 'Purchase Response was not sucessful',$hTransaction->{'successful'});

        $hReturnData = {
            'Error' => {
                'Name' => JSON_FAILURE_MESSAGES->{'Name'}->{'Purchase'},
                'Reason' => JSON_FAILURE_MESSAGES->{'Reason'}->{'ConnectionError'},
            }
        };
    }
    return $self->_writeDataToOutput($hReturnData);
}

#========================================================================================
# function     CompleteTransaction
# state        public
#----------------------------------------------------------------------------------------
# syntax       CompleteTransaction( $Servlet );
#----------------------------------------------------------------------------------------
# description  UI Controller for CompleteTransaction at multi-step checkout
#----------------------------------------------------------------------------------------
# input        $Servlet | Servlet | DE_EPAGES::Presentation::Servlet::BasePermission
#========================================================================================
sub CompleteTransaction {
    my $self = shift;
    my ($Servlet) = @_;

    my $BasketOrder = $Servlet->object();
    my $TransactionToken = $Servlet->params()->{'TransactionToken'};

    my $hTransaction;
    my $hResult;
    my $IsThreeDSAvailable = $self->_isThreeDSAvailable($BasketOrder);

    if (!$IsThreeDSAvailable) {
        $hResult = {
            'Error' => {
                'Name' => JSON_FAILURE_MESSAGES->{'Name'}->{'Complete'},
                'Reason' => JSON_FAILURE_MESSAGES->{'Reason'}->{'CardDeclined'},
            }
        };
        $self->_LogPayment($Servlet, 'Trying to call Complete Transaction on non 3DS Gateway', $hResult);
        return $self->_writeDataToOutput($hResult);
    }

    eval { $hTransaction = $self->_completeTransaction($BasketOrder, $TransactionToken); };
    if ( ExistsError() ) {
        $hResult = $self->_HandleCriticalErrorOnRequest($BasketOrder, $Servlet, 'Complete Transaction Error');
        return $self->_writeDataToOutput($hResult);
    }

    $self->_LogPayment($Servlet, 'Complete transaction', { 'Basket' =>  $BasketOrder->guid(), 'Transaction' => $hTransaction });

    if ( $hTransaction->{'successful'} ) {

        if ($hTransaction->{'state'} eq PURCHASE_RESPONSE_STATE_SUCEEDED) {
            $self->_SaveTransactionNumber($hTransaction, $BasketOrder);
        }
        $hResult = $self->_generateFormattedTransaction($hTransaction);

    } else {
        $self->_LogPayment($Servlet, 'Complete Transaction was not sucessful',$hTransaction->{'successful'});
        $hResult = {
            'Error' => {
                'Name' => JSON_FAILURE_MESSAGES->{'Name'}->{'Complete'},
                'Reason' => JSON_FAILURE_MESSAGES->{'Reason'}->{'CardDeclined'},
            }
        };
    }

    return $self->_writeDataToOutput($hResult);
}

#----------------------------------------------------------------------------------------
sub _completeTransaction {
    my $self = shift;
    my ($BasketOrder, $TransactionToken) = @_;

    my $Shop        = $BasketOrder->getSite();
    my $GatewayName = $self->_getGatewayName($BasketOrder);
    my $PurchaseController = DE_EPAGES::Spreedly::API::PurchaseController->new();
    my $hTransaction = $PurchaseController->CompleteTransaction($Shop, $GatewayName, $TransactionToken);

    return $hTransaction;
}
#----------------------------------------------------------------------------------------
sub _HandleCriticalErrorOnRequest {
    my $self = shift;
    my ($Basket, $Servlet, $LogMessage) = @_;

    my $Error = GetError();

    my $Response = $Error->vars->{Response};

    my $Reason = JSON_FAILURE_MESSAGES->{'Reason'}->{'UnknownError'};
    my $ErrorMessage = '';

    # Not an empty string and json format
    if(!$Response =~ /^$/ && $Response =~ /^\{/) {
        my $hResponse = decode_json($Response);

        $Reason = $hResponse->{'code'};
        $ErrorMessage = $hResponse->{'message'};
    }

    $self->_LogPayment($Servlet, $LogMessage,
        {
            'Basket' => $Basket->guid(),
            'ErrorMessageEp6' => $Error->message,
            'ErrorMessageEpJ' => $ErrorMessage,
        }
    );

    my $hReturnData = {
        'Error' => {
            'Name' => JSON_FAILURE_MESSAGES->{'Name'}->{'Purchase'},
            'Reason' => $Reason
        }
    };

    return $hReturnData;
}

#----------------------------------------------------------------------------------------
sub _HandleSuccessfulPurchaseResponse {
    my $self = shift;
    my ($hTransaction, $Basket, $Servlet) = @_;

    my $hReturnData;

    if ($hTransaction->{'state'} eq PURCHASE_RESPONSE_STATE_SUCEEDED) {
        $self->_SaveTransactionNumber( $hTransaction, $Basket );
        $hReturnData = {
            'Transaction' =>  {
                'state' => $hTransaction->{'state'},
                'required_action' => $hTransaction->{'required_action'},
            }
        };

    } elsif ($hTransaction->{'state'} eq PURCHASE_RESPONSE_STATE_PENDING && $self->_isThreeDSAvailable($Basket)) {
        $hReturnData = $self->_generateFormattedTransaction($hTransaction);

        # When we have a 3DS1 fallback there is no transaction number in the showTransaction response so we save it here
        if ($hTransaction->{'required_action'} eq PURCHASE_RESPONSE_REQUIRED_ACTION_REDIRECT) {
            my $LineItemPaymentSpreedly = $self->_RetrieveLineItemPayment($Basket);
            $LineItemPaymentSpreedly->saveTransactionNumber($hTransaction->{'transactionNo'});
        }

    } else {
        $self->_LogPayment($Servlet, 'Purchase Response - value of state unknown or not 3DS and pending',$hTransaction);
        $hReturnData = {
            'Error' => {
                'Name' => JSON_FAILURE_MESSAGES->{'Name'}->{'Purchase'},
                'Reason' => JSON_FAILURE_MESSAGES->{'Reason'}->{'CardDeclined'},
            }
        };
    }

    return $hReturnData;
}

#========================================================================================
# function     _isThreeDSAvailable
# state        private
#----------------------------------------------------------------------------------------
# syntax       _isThreeDSAvailable();
#----------------------------------------------------------------------------------------
# description  Returns a boolean regarding the availability of 3DS on the gateway.
#----------------------------------------------------------------------------------------
# input       $IsThreeDSAvailable  | ThreeDS availability | Boolean
#========================================================================================
sub _isThreeDSAvailable {
    my $self = shift;
    my ( $Basket ) = @_;

    my $GatewayName = $self->_getGatewayName($Basket);
    if (!isValid($GatewayName)) {
        return 0;
    }
    if ($GatewayName ~~ THREE_DS_AVAILABLE_GATEWAYS) {
        return 1;
    }

    return 0;
}
#----------------------------------------------------------------------------------------
sub _SaveTransactionToken {
    my $self = shift;
    my ( $TransactionToken, $LineItemContainer ) = @_;

    my $LineItemPaymentSpreedly = $LineItemContainer->get('Payment');
    $LineItemPaymentSpreedly->setTransactionToken($TransactionToken);

    return;
}

#----------------------------------------------------------------------------------------
sub _SaveTransactionNumber {
    my $self = shift;
    my ( $hTransaction, $Basket ) = @_;

    my $LineItemContainer       = $Basket->get('LineItemContainer');
    my $LineItemPaymentSpreedly = $LineItemContainer->get('Payment');
    $LineItemPaymentSpreedly->updateTransactionStatus(PAYMENT_STATUS_PAID, $hTransaction);

    return;
}

#----------------------------------------------------------------------------------------
# Wrappers
#----------------------------------------------------------------------------------------
sub _getGatewayName {
    my $self = shift;
    my ( $Basket ) = @_;

    my $LineItemContainer = $Basket->get('LineItemContainer');
    return undef unless isValid($LineItemContainer);

    my $LineItemPaymentSpreedly = $LineItemContainer->get('Payment');
    return undef unless isValid($LineItemPaymentSpreedly);

    my $PaymentMethod = $LineItemPaymentSpreedly->get('PaymentMethod');
    return undef unless isValid($PaymentMethod);

    my $PaymentType = $PaymentMethod->get('PaymentType');
    return undef unless isValid($PaymentType);

    return $PaymentType->get('GatewayName');
}

#----------------------------------------------------------------------------------------
sub _LogPayment {
    my $self = shift;
    my ($Servlet, $Name, $DebugStuff) = @_;

    my $Basket = $Servlet->object();
    my $LineItemPayment = $self->_RetrieveLineItemPayment($Basket);
    my $PaymentMethod   = $LineItemPayment->get('PaymentMethod');
    my $PaymentTypeAlias = $PaymentMethod->get('PaymentType')->alias();

    LogPayment($PaymentTypeAlias, $Name, $DebugStuff);

    return;
}

#----------------------------------------------------------------------------------------
sub _makePurchase {
    my $self = shift;
    my ( $LineItemContainer, $Basket, $PurchaseParams) = @_;

    my $Shop             = $Basket->getSite();
    my $GatewayName = $self->_getGatewayName($Basket);
    my $PurchaseController = DE_EPAGES::Spreedly::API::PurchaseController->new();

    return $PurchaseController->makePurchase( $LineItemContainer, $Shop, $GatewayName, $PurchaseParams);
}

#----------------------------------------------------------------------------------------
sub _RetrieveLineItemPayment {
    my $self = shift;
    my ($Basket) = @_;

    my $LineItemContainer = $Basket->container();

    return $LineItemContainer->get('Payment');
}

#----------------------------------------------------------------------------------------
sub _writeDataToOutput {
    my $self = shift;
    my ($hData) = @_;

    $self->writeContentOutput($hData);
    return;
}

#----------------------------------------------------------------------------------------
sub _getRedirectUrl {
    my $self = shift;
    my ($Servlet, $Basket, $hServletParams) = @_;

    my $hSuccessUrlParams = {
        'ViewAction' => 'ViewMultiCheckoutConfirmation',
    };
    my $hCallbackUrlOptions = {
        'Type' => 'sf/mobile',
        'AddToken' => '1',
        'UseSSL' => '1',
    };
    my $SuccessUrl = $self->_getCallbackShopUrl($Basket, $hSuccessUrlParams, $hCallbackUrlOptions, $Servlet);

    if ($hServletParams->{'IsRatingInvitationAllowed'}) {
        return sprintf('%s&IsRatingInvitationAllowed=1', $SuccessUrl);
    }

    return $SuccessUrl;
}

#----------------------------------------------------------------------------------------
sub _getCallbackUrl {
    my $self = shift;
    my ($Servlet, $Basket) = @_;

    my $Shop = $Basket->get('Shop');

    my $PaymentLineItemGuid = $Basket->container()->get('Payment')->guid();

    my $hParams = {
        'ChangeAction' => 'SpreedlyPaymentCallback',
        'PaymentLineItemGUID' => $PaymentLineItemGuid
    };

    my $hOptions = { 'Type' => 'sf/mobile', 'UseSSL' => 1};
    my $URL = $self->_getCallbackShopUrl( $Shop, $hParams, $hOptions, $Servlet );

    my $PayLoad = sprintf('%s&%s&%s', $Shop->alias(), $Shop->id(), $PaymentLineItemGuid);
    my $DereferrerToken = $self->_GenerateDereferrerToken($PayLoad);

    return sprintf('%s&DereferrerToken=%s', $URL, $DereferrerToken);
}

#----------------------------------------------------------------------------------------
sub _getCallbackShopUrl {
    my $self = shift;
    my ($Object, $hArgs, $hOptions, $Servlet) = @_;

    return BuildShopUrl($Object, $hArgs, $hOptions, $Servlet);
}

#========================================================================================
# function     _GenerateDereferrerToken
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Token = $self->_GenerateDereferrerToken($URL);
#----------------------------------------------------------------------------------------
# description  Wrapper for GenerateDereferrerToken
#----------------------------------------------------------------------------------------
# input        $URL  | url | string
# input        $TimeStamp  | unix timestamp (optional) | Integer
#----------------------------------------------------------------------------------------
# return       $DereferrerToken | base 64 encoded token | String
#========================================================================================
sub _GenerateDereferrerToken {
    my $self = shift;
    my ( $URL, $TimeStamp ) = @_;
    return GenerateDereferrerToken( $URL, $TimeStamp );
}

#----------------------------------------------------------------------------------------
sub _generateFormattedTransaction {
    my $self = shift;
    my ($hTransaction) = @_;

    return {
        'Transaction' =>  {
            'state' => $hTransaction->{'state'},
            'token' => $hTransaction->{'token'},
            'required_action' => $hTransaction->{'required_action'},
            'device_fingerprint_form' => $hTransaction->{'device_fingerprint_form'},
            'challenge_url' => $hTransaction->{'challenge_url'},
            'challenge_form' => $hTransaction->{'challenge_form'},
            'checkout_form' => $hTransaction->{'checkout_form'},
            'checkout_url' => $hTransaction->{'checkout_url'},
        }
    };
}

#----------------------------------------------------------------------------------------
sub _validateGatewayRequiredFields {
    my $self = shift;
    my ($PurchaseParams, $Basket, $Servlet) = @_;

    my $GatewayName = $self->_getGatewayName($Basket);
    my $BillingAddress = $Basket->get('BillingAddress');
    my $BillingCountryCode2 = $BillingAddress->get('Country')->{'Code2'};

    # payu_latam mexico - birthday and device session id are required
    if($GatewayName eq SPREEDLY_GATEWAY_NAME_PAYU_LATAM &&
        $BillingCountryCode2 eq COUNTRY_CODE2_MX) {
        my $hResult = $self->_validateBillingAddress( $Basket, $Servlet );
        if(defined($hResult->{'isError'}) && $hResult->{'isError'} == 1) {
            Error('GatewayRequiredFields error. '. $hResult->{'message'});
        }

        my $GatewayFieldBirthday = $Servlet->params('GatewayFieldBirthday');
        if(isValid($GatewayFieldBirthday)) {
            $PurchaseParams->{'GatewayFieldBirthday'} = $GatewayFieldBirthday;
        }

        my $GatewayFieldDeviceSessionId = $self->_createPayUDeviceSessionId($Servlet->cookies('SessionID'));
        if(isValid($GatewayFieldDeviceSessionId)) {
            $PurchaseParams->{'GatewayFieldDeviceSessionId'} = $GatewayFieldDeviceSessionId;
        }

    }

    return;
}

#----------------------------------------------------------------------------------------
sub _validateBillingAddress {
    my $self = shift;
    my ($Basket, $Servlet) = @_;
    my $hResult = { 'isError' => 0, 'message' => "" };

    my $BillingAddress = $Basket->get('BillingAddress');
    if(not defined $BillingAddress) {
        $hResult = { 'isError' => 1, 'message' => "Invalid BillingAddress" };
        return $hResult;
    }

    my $BillingCountryCode2 = $BillingAddress->get('Country')->{'Code2'};
    if(not defined $BillingCountryCode2) {
        $hResult = { 'isError' => 1, 'message' => "Invalid BillingAddress BillingCountryCode2" };
        return $hResult;
    }

    # birthday is mandatory for mexican transactions
    if($BillingCountryCode2 eq COUNTRY_CODE2_MX) {
        my $BillingAddressBirthday = $BillingAddress->get('Birthday');
        my $GatewayFieldBirthday = $Servlet->params()->{'GatewayFieldBirthday'};
        if(isValid($BillingAddressBirthday)) {
            $hResult = { 'isError' => 0, 'message' => "BillingAddress Birthday" };
        } else {
            if(isValid($GatewayFieldBirthday)) {
                $hResult = { 'isError' => 0, 'message' => "GatewayField Birthday" };
            } else {
                $hResult = { 'isError' => 1, 'message' => "Error invalid field: Birthday. The Birthday is mandatory for transactions in Mexico." };
            }
        }
    }

    return $hResult;
}

#----------------------------------------------------------------------------------------
sub _createPayUDeviceSessionId {
    my $self = shift;
    my ($SessionID) = @_;

    return undef unless defined $SessionID;

    (my $EpochSeconds, my $Microseconds) = gettimeofday();
    my $Microtime = sprintf("%s %s", $EpochSeconds, $Microseconds);
    my $DeviceSessionId = md5_hex($SessionID.$Microtime);

    return $DeviceSessionId;
}

1;
